"""Test hdf5 file format"""
import io
import os
import pathlib
import shutil
import sys
import tempfile

import h5py
import numpy as np
import pytest

import dclab
from dclab import new_dataset, rtdc_dataset
from dclab.rtdc_dataset import config, rtdc_copy

from helper_methods import retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_config():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert ds.config["setup"]["channel width"] == 30
    assert ds.config["setup"]["chip region"].lower() == "channel"
    assert ds.config["setup"]["flow rate"] == 0.16
    assert ds.config["imaging"]["pixel size"] == 0.34


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_contour_basic():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert len(ds) == 5
    assert len(ds["contour"]) == 5
    assert ds["deform"].ndim == 1  # important for matplotlib
    assert np.allclose(np.average(ds["contour"][0]), 30.75)
    assert np.median(ds["contour"][0]) == 30.5
    assert np.median(ds["contour"][2]) == 31.5
    assert np.median(ds["contour"][-3]) == 31.5
    assert np.median(ds["contour"][4]) == 33.0
    assert np.median(ds["contour"][-1]) == 33.0


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_defective_feature_aspect():
    # see https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut/issues/241
    h5path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # modify aspect feature
    with h5py.File(h5path, "a") as h5:
        aspect0 = h5["events/aspect"][0]
        aspect1 = 1.234
        assert not np.allclose(aspect0, aspect1), "test's sanity"
        h5["events/aspect"][0] = aspect1
        # In Shape-In 2.0.5 everything worked fine
        h5.attrs["setup:software version"] = "ShapeIn 2.0.5"
    # sanity check
    with new_dataset(h5path) as ds1:
        assert np.allclose(ds1["aspect"][0], aspect1)
    # trigger recomputation of aspect feature
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "ShapeIn 2.0.6"
    # verify original value of aspect
    with new_dataset(h5path) as ds2:
        assert np.allclose(ds2["aspect"][0], aspect0)


@pytest.mark.parametrize("feat", [
    "inert_ratio_cvx",
    "inert_ratio_prnc",
    "inert_ratio_raw",
    "tilt"])
def test_defective_feature_inert_ratio_prnc(feat):
    # see https://github.com/DC-analysis/dclab/issues/212
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    # sanity check
    with h5py.File(h5path, "a") as h5:
        size = len(h5["/events/deform"])
        h5.attrs["setup:software version"] = "dclab 0.48.0"
        h5[f"/events/{feat}"] = np.ones(size)

    tmp_path = pathlib.Path(tempfile.mkdtemp("test_212")) / "test_212.rtdc"

    with new_dataset(h5path) as ds:
        # "feat" is defective
        assert feat not in ds.features_innate
        ds.export.hdf5(tmp_path, features=["deform", feat])

    with new_dataset(tmp_path) as ds:
        assert feat in ds.features_innate
        assert not np.all(ds[feat] == 1)


def test_defective_feature_inert_ratio_control_1():
    feat = "inert_ratio_prnc"
    # see https://github.com/DC-analysis/dclab/issues/212
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    # sanity check
    with h5py.File(h5path, "a") as h5:
        size = len(h5["/events/deform"])
        h5.attrs["setup:software version"] = "dclab 0.48.3"
        h5[f"/events/{feat}"] = np.ones(size)

    with new_dataset(h5path) as ds:
        # "feat" is defective
        assert feat in ds.features_innate
        assert np.all(ds[feat] == 1)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_defective_feature_inert_ratio_control_2():
    feat = "inert_ratio_prnc"
    # see https://github.com/DC-analysis/dclab/issues/212
    h5path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    # sanity check
    with h5py.File(h5path, "a") as h5:
        size = len(h5["/events/deform"])
        h5.attrs["setup:software version"] = "dclab 0.48.0"
        h5[f"/events/{feat}"] = np.ones(size)

    with new_dataset(h5path) as ds:
        # "feat" is defective
        assert feat in ds.features_innate
        assert np.all(ds[feat] == 1)


def test_defective_feature_time_issue_204_float32():
    # see https://github.com/DC-analysis/dclab/issues/204
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    # sanity check
    with h5py.File(h5path, "r") as h5:
        assert h5["events/time"].dtype.char == "f"

    with new_dataset(h5path) as ds:
        # "time" is a rapid feature, so it will be in features_innate
        assert "time" not in ds.features_innate


def test_defective_feature_time_issue_204_float32_exported():
    # see https://github.com/DC-analysis/dclab/issues/204
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    # sanity check
    with h5py.File(h5path, "r") as h5:
        assert h5["events/time"].dtype.char == "f"

    tmp_path = pathlib.Path(tempfile.mkdtemp("test_204")) / "test_204.rtdc"

    with new_dataset(h5path) as ds:
        # "time" is a rapid feature, so it will be in features_innate
        ds.export.hdf5(tmp_path, features=["deform", "time"])

    with new_dataset(tmp_path) as ds:
        # time should now be innate
        assert "time" in ds.features_innate


def test_defective_feature_time_issue_204_noancil():
    # see https://github.com/DC-analysis/dclab/issues/204
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    # sanity check
    with h5py.File(h5path, "a") as h5:
        assert h5["events/time"].dtype.char == "f"
        del h5["events/frame"]

    with new_dataset(h5path) as ds:
        assert "time" in ds.features_innate  # now it's there


def test_defective_feature_time_issue_204_noshapein():
    # see https://github.com/DC-analysis/dclab/issues/204
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    # sanity check
    with h5py.File(h5path, "a") as h5:
        assert h5["events/time"].dtype.char == "f"
        time = np.arange(h5["events/time"].size, dtype=float)
        del h5["events/time"]
        h5["events/time"] = time
        # Set the software version to something else than Shape-In, but with
        # a lower version of dclab.
        h5.attrs["setup:software version"] = "Other device | dclab 0.30.0"

    with new_dataset(h5path) as ds:
        assert "time" in ds.features_innate  # It's not Shape-In, thus ignored


def test_defective_feature_time_issue_204_old_dclab():
    # see https://github.com/DC-analysis/dclab/issues/204
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    # sanity check
    with h5py.File(h5path, "a") as h5:
        assert h5["events/time"].dtype.char == "f"
        time = np.arange(h5["events/time"].size, dtype=float)
        del h5["events/time"]
        h5["events/time"] = time
        # Set the software version to something else than Shape-In, but with
        # a lower version of dclab.
        h5.attrs["setup:software version"] = "ShapeIn 2.0.2 | dclab 0.30.0"

    with new_dataset(h5path) as ds:
        assert "time" not in ds.features_innate  # Recognized as old dclab


def test_defective_feature_time_issue_204_new_dclab():
    # see https://github.com/DC-analysis/dclab/issues/204
    h5path = retrieve_data("fmt-hdf5_polygon_gate_2021.zip")
    # sanity check
    with h5py.File(h5path, "a") as h5:
        assert h5["events/time"].dtype.char == "f"
        time = np.arange(h5["events/time"].size, dtype=float)
        del h5["events/time"]
        h5["events/time"] = time
        # Set the software version to something else than Shape-In, but with
        # a lower version of dclab.
        h5.attrs["setup:software version"] = "ShapeIn 2.0.2 | dclab 0.50.0"

    with new_dataset(h5path) as ds:
        assert "time" in ds.features_innate  # Recognized as new dclab


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_defective_feature_time_issue_207_with_offset():
    # see https://github.com/DC-analysis/dclab/issues/207
    h5path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # sanity check
    with h5py.File(h5path, "a") as h5:
        h5.attrs["imaging:frame rate"] = 100
        # write modified frames
        frame_w_offset = h5["events/frame"][:]
        frame_w_offset += 42 - frame_w_offset[0]
        del h5["events/frame"]
        h5["events/frame"] = frame_w_offset
        # write wrong time
        bad_time = np.arange(h5["events/time"].size, dtype=np.float32)
        del h5["events/time"]
        h5["events/time"] = bad_time

    with new_dataset(h5path) as ds:
        assert "time" not in ds.features_innate  # Recognized as float32
        assert np.allclose(ds["time"][0], 0.42, atol=0, rtol=1e-7)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_defective_feature_volume():
    # see https://github.com/ZELLMECHANIK-DRESDEN/ShapeOut/issues/241
    h5path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")

    with h5py.File(h5path, "r") as h5:
        # This volume was computed with an old version of dclab
        wrong_volume = h5["events/volume"][:]

    with new_dataset(h5path) as ds:
        assert "volume" not in ds.features_innate
        correct_volume = ds["volume"]
        assert not np.allclose(wrong_volume, correct_volume)

    # prevent recomputation via logs
    # (do not use context manager here [sic])
    hw = dclab.RTDCWriter(h5path)
    hw.store_log("dclab_issue_141", ["fixed"])
    hw.h5file.close()
    with new_dataset(h5path) as ds2:
        assert np.allclose(ds2["volume"], wrong_volume)

    # remove the log
    with h5py.File(h5path, "a") as h5:
        del h5["logs/dclab_issue_141"]

    # make sure that worked
    with new_dataset(h5path) as ds:
        assert "volume" not in ds.features_innate
        correct_volume = ds["volume"]
        assert not np.allclose(wrong_volume, correct_volume)

    # prevent recomputation via dclab version string
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "ShapeIn 2.0.6 | dclab 0.37.0"
    with new_dataset(h5path) as ds2:
        assert "volume" in ds2.features_innate
        assert np.allclose(ds2["volume"], wrong_volume)

    # reset version string
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "ShapeIn 2.0.6 | dclab 0.35.1"
    with new_dataset(h5path) as ds2:
        assert "volume" not in ds2.features_innate
        assert np.allclose(ds2["volume"], correct_volume)

    # use context manager to write version number
    with dclab.RTDCWriter(h5path) as hw:
        pass
    with new_dataset(h5path) as ds2:
        assert "volume" in ds2.features_innate
        assert np.allclose(ds2["volume"], wrong_volume)


def test_discouraged_array_dunder_childndarray():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip"))
    with pytest.warns(UserWarning, match="It may consume a lot of memory"):
        ds["mask"].__array__()


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_hash():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert ds.hash == "2c436daba22d2c7397b74d53d80f8931"
    assert ds.format == "hdf5"


def test_fileio_basic():
    """Basic file IO functionality tests"""
    path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    # Create a file object
    fd = io.BytesIO()
    # create a copy
    with h5py.File(fd, "w") as hio, h5py.File(path) as h5:
        rtdc_copy(src_h5file=h5, dst_h5file=hio)

    ds = dclab.new_dataset(fd)
    assert len(ds) == 10
    assert str(ds).count("RTDC_HDF5")
    assert ds.__repr__().count("mm-hdf5")
    assert ds.__repr__().count("BytesIO")
    assert np.allclose(ds["deform"].mean(),
                       0.03620073944330215454,
                       atol=0, rtol=1e-10)
    assert np.allclose(ds.h5file["/events/deform"].attrs["mean"],
                       0.03620073944330215454,
                       atol=0, rtol=1e-10)
    assert "volume" in ds.features
    assert "volume" not in ds.features_innate
    assert np.allclose(ds["volume"][0], 2898.67525650645,
                       rtol=1e-6, atol=0)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_load_non_scalar_data():
    """Loading non-scalar data should return data as h5py.Dataset"""
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert isinstance(ds["image"], h5py.Dataset)
    assert isinstance(ds["image_bg"], h5py.Dataset)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_shape_contour():
    ds = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    assert "contour" in ds.features_innate
    assert len(ds["contour"]) == 8
    assert ds["contour"].shape == (8, np.nan, 2)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_shape_image():
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert "image" in ds.features_innate
    assert len(ds["image"]) == 5
    assert ds["image"].shape == (5, 80, 250)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_shape_mask():
    ds = new_dataset(retrieve_data("fmt-hdf5_image-bg_2020.zip"))
    assert "mask" in ds.features_innate
    assert len(ds["mask"]) == 5
    assert ds["mask"].shape == (5, 80, 250)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hdf5_shape_trace():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2018.zip"))
    assert len(ds) == 7
    assert "trace" in ds.features_innate
    assert ds["trace"].shape == (6, 7, 177)
    assert ds["trace"]["fl1_raw"].shape == (7, 177)
    assert ds["trace"]["fl1_raw"][0].shape == (177,)
    assert len(ds["trace"]) == 6
    assert len(ds["trace"]["fl1_raw"]) == 7
    assert len(ds["trace"]["fl1_raw"][0]) == 177


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.skipif(sys.version_info < (3, 9),
                    reason="requires python3.9 or higher")
def test_hdf5_ufuncs():
    path_orig = retrieve_data("fmt-hdf5_fl_2018.zip")
    path_mod = path_orig.with_stem("modified")
    shutil.copy2(path_orig, path_mod)
    with h5py.File(path_mod, "a") as h5:
        h5["events/area_cvx"].attrs["min"] = 10.
        h5["events/area_cvx"].attrs["max"] = 100.
        h5["events/area_cvx"].attrs["mean"] = 12.7

    ds = new_dataset(path_orig)
    ds_mod = new_dataset(path_mod)

    assert len(ds) == 7
    assert len(ds_mod) == 7

    # modified
    assert np.min(ds_mod["area_cvx"]) == 10
    assert np.max(ds_mod["area_cvx"]) == 100
    assert np.mean(ds_mod["area_cvx"]) == 12.7

    # reference
    assert np.min(ds["area_cvx"]) == 226.0
    assert np.max(ds["area_cvx"]) == 287.5
    assert np.allclose(np.mean(ds["area_cvx"]), 255.28572, rtol=0, atol=1e-5)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ignore_empty_hdf5_meta_data_attribute():
    """Ignore empty hdf5 attributes / dclab metadata"""
    # see https://github.com/DC-analysis/dclab/issues/109
    path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # add empty attribute
    with h5py.File(path, "r+") as h5:
        h5.attrs["setup:module composition"] = ""

    # assert
    with pytest.warns(config.EmptyConfigurationKeyWarning,
                      match=r"Empty value for \[setup\]: 'module composition'!"
                      ):
        new_dataset(path)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_ignore_unknown_hdf5_meta_data_attribute():
    """Ignore unknown hdf5 attributes / dclab metadata"""
    # see https://github.com/DC-analysis/dclab/issues/109
    path = retrieve_data("fmt-hdf5_fl_2018.zip")
    # add empty attribute
    with h5py.File(path, "r+") as h5:
        h5.attrs["setup:shizzle"] = ""

    # assert
    with pytest.warns(config.UnknownConfigurationKeyWarning,
                      match=r"Unknown key 'shizzle' in the 'setup' section!"
                      ):
        new_dataset(path)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_image_basic():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert np.allclose(np.average(ds["image"][1]), 125.37133333333334)
    assert len(ds["image"]) == 5


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_image_bg():
    path = retrieve_data("fmt-hdf5_fl_2017.zip")
    # add a fake image_bg column
    with rtdc_dataset.RTDCWriter(path) as hw:
        image_bg = hw.h5file["events"]["image"][:] // 2
        hw.store_feature("image_bg", image_bg)

    with new_dataset(path) as ds:
        for ii in range(len(ds)):
            assert np.all(ds["image"][ii] // 2 == ds["image_bg"][ii])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_image_bg_2():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    with new_dataset(path) as ds:
        assert "image_bg" in ds
        bgc = ds["image"][0] - ds["image_bg"][0]
        assert bgc[10, 11] == 6


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
@pytest.mark.parametrize("idxs",
                         [slice(0, 5, 2),
                          [0, 2, 4],
                          np.array([0, 2, 4]),
                          np.array([True, False, True, False, True])])
def test_index_slicing_contour(idxs):
    data = retrieve_data("fmt-hdf5_fl_2017.zip")
    ds = new_dataset(data)

    contour_ref = [
        ds["contour"][0],
        ds["contour"][2],
        ds["contour"][4],
    ]

    contour_slice = ds["contour"][idxs]

    assert np.all(contour_slice[0] == contour_ref[0])
    assert np.all(contour_slice[1] == contour_ref[1])
    assert np.all(contour_slice[2] == contour_ref[2])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_logs():
    path_in = retrieve_data("fmt-hdf5_mask-contour_2018.zip")

    with new_dataset(path_in) as ds:
        assert not ds.logs

    # write some logs
    with h5py.File(path_in, "a") as h5:
        hw = rtdc_dataset.RTDCWriter(h5)
        hw.store_log("test_log",  ["peter", "hans"])

    with new_dataset(path_in) as ds:
        assert ds.logs
        assert ds.logs["test_log"][0] == "peter"

    # remove logs
    with h5py.File(path_in, "a") as h5:
        del h5["logs"]

    with new_dataset(path_in) as ds:
        assert not ds.logs
        try:
            ds.logs["test_log"]
        except KeyError:  # no log data
            pass


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_no_suffix():
    """Loading an .rtdc file that has a wrong suffix"""
    path = str(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    path2 = path + ".wrong_suffix"
    os.rename(path, path2)
    ds = new_dataset(path2)
    assert (len(ds) == 8)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_online_polygon_filters():
    """Shape-In 2.3 supports online polygon filters"""
    path = retrieve_data("fmt-hdf5_mask-contour_2018.zip")
    # add an artificial online polygon filter
    with h5py.File(path, "a") as h5:
        # set soft filter to True
        h5.attrs["online_filter:area_um,deform soft limit"] = True
        # set filter values
        pf_name = "online_filter:area_um,deform polygon points"
        area_um = h5["events"]["area_um"]
        deform = h5["events"]["deform"]
        pf_points = np.array([
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform)],
            [np.mean(area_um) + np.std(area_um),
             np.mean(deform) + np.std(deform)],
            [np.mean(area_um),
             np.mean(deform) + np.std(deform)],
        ])
        h5.attrs[pf_name] = pf_points

    # see if we can open the file without any error
    with new_dataset(path) as ds:
        assert len(ds) == 8
        assert ds.config["online_filter"]["area_um,deform soft limit"]
        assert "area_um,deform polygon points" in ds.config["online_filter"]
        assert np.allclose(
            ds.config["online_filter"]["area_um,deform polygon points"],
            pf_points)


def test_open_simple_file(tmp_path):
    path = tmp_path / "test.rtdc"
    with h5py.File(path, "w") as h5:
        h5["events/time"] = np.arange(100)
        h5["events/deform"] = np.linspace(0.1, 0.2, 100)

    with dclab.new_dataset(path) as ds:
        assert "software version" not in ds.config["setup"]
        assert ds.title == "undefined sample - M0"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_open_with_invalid_feature_names():
    """Loading an .rtdc file that has wrong feature name"""
    path = str(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))
    # add wrong feature name right at the top of the list
    with h5py.File(path, "r+") as h5:
        h5["events"]["a0"] = h5["events"]["deform"]
    # see if we can open the file without any error
    ds = new_dataset(path)
    assert len(ds) == 8


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
def test_trace():
    ds = new_dataset(retrieve_data("fmt-hdf5_fl_2017.zip"))
    assert len(ds["trace"]) == 2
    assert ds["trace"]["fl1_raw"].shape == (5, 100)
    assert np.allclose(np.average(
        ds["trace"]["fl1_median"][0]), 0.027744706519425219)



"""Extract volume-deformation isoelasticity lines from simulation data

The simulation dataset must be an HDF5 file with a specific structure
that contains the results of FEM simulations. The original HDF5 files
were provided by Lucas Wittwer.

Creating volume-deformation isoelasticity lines means that a full LUT
is generated from which the isoelasticity lines are then interpolated.
Since the LUT has no function in dclab, it is not exported.

Additional post-processing hooks for LUT or isoelastics generation
are defined in the Python files named according to the LUT identifier
in the "lut_recipes" subdirectory.

The discussion related to this script is archived in issue #70 (dclab).

An example HDF5 file can be found on figshare
(LE-2D-FEM-19, https://doi.org/10.6084/m9.figshare.12155064.v4).
"""
import argparse
import pathlib

import matplotlib.pylab as plt

from lut_recipes import LutProcessor
import fem2lutiso_std


def main():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('input', metavar="INPUT", type=str,
                        help='Input path (.hdf5 file)')
    parser.add_argument("--raw",
                        dest='raw',
                        action='store_true',
                        help="do not perform data post-processing",
                        )
    parser.set_defaults(raw=False)

    args = parser.parse_args()
    path = pathlib.Path(args.input)
    raw = args.raw

    if raw:
        print("Skipping all post-processing steps!")

    print("Extracting volume-deformation LUT")
    lup = LutProcessor(path, use_hooks=not raw, featx="volume", verbose=True)
    lut, contours, levels = lup.assemble_lut_and_isoelastics()

    ax = plt.subplot(111, title="Final LUT and isoelastics")
    ax.plot(lut[:, 0], lut[:, 1], ".", color="k")
    for cc in contours:
        ax.plot(cc[:, 0], cc[:, 1])
    plt.show()

    fem2lutiso_std.save_iso(
        path=path.with_name(path.name.rsplit(".", 1)[0] + "_volume_iso.txt"),
        contours=contours,
        levels=levels,
        meta=lup.meta,
        header=["volume [um^3]", "deform", "emodulus [kPa]"])


if __name__ == "__main__":
    main()



"""Tests for mapped basins

A mapped basin is a basins that contains more events than the dataset that
is referring to the basin. A mapped basin must contain all events that are
defined in the referring dataset.
"""
import json
import sys
import uuid

import h5py
import numpy as np

import pytest

import dclab
from dclab import new_dataset, RTDCWriter
from dclab.rtdc_dataset.feat_basin import (
    BasinmapFeatureMissingError, basin_priority_sorted_key
)


from helper_methods import DCOR_AVAILABLE, retrieve_data


class recursionlimit:
    def __init__(self, limit):
        self.limit = limit

    def __enter__(self):
        self.old_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(self.limit)

    def __exit__(self, type, value, tb):
        sys.setrecursionlimit(self.old_limit)


def test_basin_basic_inception():
    """Create a mapped basin of a mapped basin"""
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    with h5py.File(h5path, "a") as h5:
        # delete circularity to avoid ancillary feature computation in this
        # test.
        del h5["events"]["circ"]

    h5path_l1 = h5path.with_name("level1.rtdc")
    basin_map1 = np.array([1, 7, 10, 14], dtype=np.uint64)
    h5path_l2 = h5path.with_name("level2.rtdc")
    basin_map2 = np.array([1, 2], dtype=np.uint64)

    # level 1
    with dclab.new_dataset(h5path) as ds0, dclab.RTDCWriter(h5path_l1) as hw1:
        hw1.store_metadata(ds0.config.as_dict(pop_filtering=True))
        hw1.store_basin(basin_name="level1",
                        basin_type="file",
                        basin_format="hdf5",
                        basin_locs=[h5path],
                        basin_map=basin_map1
                        )

    # level 2
    with dclab.new_dataset(h5path) as ds0, dclab.RTDCWriter(h5path_l2) as hw2:
        hw2.store_metadata(ds0.config.as_dict(pop_filtering=True))
        hw2.store_basin(basin_name="level2",
                        basin_type="file",
                        basin_format="hdf5",
                        basin_locs=[h5path_l1],
                        # Explicitly name the basin here, otherwise the
                        # writer will override `basinmap0` with no `basinmap0`
                        # defined in h5path_l2.
                        basin_map=("basinmap2", basin_map2)
                        )

    # Checks for level 1
    with dclab.new_dataset(h5path) as ds0, dclab.new_dataset(h5path_l1) as ds1:
        assert np.all(ds1["basinmap0"] == basin_map1)
        assert len(ds1.basins) == 1
        assert ds1.basins[0].verify_basin()
        assert "mapped" in str(ds1.basins[0])
        assert "deform" in ds1.basins[0].features
        assert np.all(ds1["deform"] == ds0["deform"][basin_map1])

    # Checks for level 2
    with dclab.new_dataset(h5path) as ds0, dclab.new_dataset(h5path_l2) as ds2:
        assert np.all(ds2["basinmap2"] == basin_map2)
        assert len(ds2.basins) == 1
        assert "deform" in ds2.basins[0].features
        assert np.all(ds2["deform"] == ds0["deform"][basin_map1][basin_map2])


def test_basin_inception():
    """Make sure an error is thrown when diving into a deep inception

    We have ten basinmap features defined in dclab. Nobody will probably
    ever exhaust them (hello future self from the past :wave:), but just
    in case this happens, an informative error should be raised.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    h5path_incept = h5path.with_name("inception.rtdc")
    mapping_array = np.arange(10, dtype=np.uint64)

    with h5py.File(h5path) as h5:
        attrs = dict(h5.attrs)

    with dclab.RTDCWriter(h5path_incept) as hw:
        hw.h5file.attrs.update(attrs)
        # Fill up all basinmap features
        for ii in range(10):
            # create a file-based basin with mapped content
            hw.store_feature(f"basinmap{ii}", mapping_array)
        # attempt to store a basin with a different basin mapping
        with pytest.raises(ValueError,
                           match="You have exhausted the usage of mapped "
                                 "basins"):
            hw.store_basin(
                basin_name="overload",
                basin_locs=[h5path],
                basin_format="hdf5",
                basin_type="file",
                basin_map=mapping_array + 1,  # has to be different
            )


@pytest.mark.parametrize("basinmap", [
    [1, 3, 4],  # very simple case
    [1, 1, 1, 2],  # not trivial, not realizable with hierarchy children
])
def test_basin_mapped(basinmap):
    path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    with h5py.File(path, "a") as h5:
        # delete circularity to avoid ancillary feature computation in this
        # test.
        del h5["events"]["circ"]

    path_out = path.with_name("level1.rtdc")
    basinmap = np.array(basinmap, dtype=np.uint64)

    # create basin
    with dclab.new_dataset(path) as ds0, dclab.RTDCWriter(path_out) as hw1:
        hw1.store_metadata(ds0.config.as_dict(pop_filtering=True))
        hw1.store_basin(basin_name="level1",
                        basin_type="file",
                        basin_format="hdf5",
                        basin_locs=[path],
                        basin_map=basinmap
                        )

    # Checks basin
    with dclab.new_dataset(path) as ds0, dclab.new_dataset(path_out) as ds1:
        assert np.all(ds1["basinmap0"] == basinmap)
        assert len(ds1.basins) == 1
        assert ds1.basins[0].verify_basin()
        assert "mapped" in str(ds1.basins[0])
        assert "deform" in ds1.basins[0].features
        assert np.all(ds1["deform"] == ds0["deform"][basinmap])
        assert np.all(ds1["image"] == ds0["image"][:][basinmap])
        assert np.all(ds1["mask"] == ds0["mask"][:][basinmap])


@pytest.mark.parametrize("basinmap", [
    [1, 3, 4],  # very simple case
    [1, 1, 1, 2],  # not trivial, not realizable with hierarchy children
])
@pytest.mark.filterwarnings(
    "ignore::UserWarning")
def test_basin_mapped_export(basinmap):
    path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    with h5py.File(path, "a") as h5:
        # delete circularity to avoid ancillary feature computation in this
        # test.
        del h5["events"]["circ"]

    path_out = path.with_name("level1.rtdc")
    basinmap = np.array(basinmap, dtype=np.uint64)

    # create basin
    with dclab.new_dataset(path) as ds0, dclab.RTDCWriter(path_out) as hw1:
        hw1.store_metadata(ds0.config.as_dict(pop_filtering=True))
        hw1.store_basin(basin_name="level1",
                        basin_type="file",
                        basin_format="hdf5",
                        basin_locs=[path],
                        basin_map=basinmap
                        )

    path_export = path.with_name("export.rtdc")
    with dclab.new_dataset(path_out) as ds1:
        ds1.filter.manual[:] = False
        ds1.filter.manual[:2] = True
        ds1.apply_filter()
        ds1.export.hdf5(
            path_export,
            features=["deform", "image", "mask"]
        )

    # Export data
    with dclab.new_dataset(path) as ds0, dclab.new_dataset(path_export) as ds1:
        assert np.all(ds1["deform"] == ds0["deform"][basinmap[:2]])
        assert np.all(ds1["image"] == ds0["image"][:][basinmap[:2]])
        assert np.all(ds1["mask"] == ds0["mask"][:][basinmap[:2]])


def test_error_when_basinmap_not_given():
    """This is a test for when the basinmap feature for mapping is missing

    A RecursionError is raised by dclab if the basinmap feature
    cannot be found in the data.
    """
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    # create a file-based basin with mapped content
    h5path_small = h5path.with_name("smaller.rtdc")
    mapping_array = np.array([1, 7, 10, 14], dtype=np.uint64)

    # Dataset creation
    with h5py.File(h5path, "a") as src, RTDCWriter(h5path_small) as hw:
        # write experiment identifiers
        hw.h5file.attrs.update(src.attrs)
        unique_id = f"unique_id_{uuid.uuid4()}"
        src.attrs["experiment:run identifier"] = unique_id
        hw.h5file.attrs["experiment:run identifier"] = unique_id + "-hans"

        assert src["events/deform"].size == 18, "sanity check"
        assert "area_um" in src["events"]

        # first, copy parts of the scalar features to the new file
        hw.store_feature("deform", src["events/deform"][mapping_array])

        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "features": ["area_um"],
            "paths": [
                str(h5path),  # absolute path name
            ],
            "mapping": "basinmap0"  # note that this feature does not exist
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "invalid_format_basin", blines)

    with dclab.new_dataset(h5path) as ds0, \
            dclab.new_dataset(h5path_small) as dss:
        # This is the feature that is in the small file
        assert np.all(ds0["deform"][mapping_array] == dss["deform"])
        # The area_um feature cannot be retrieved from the original file,
        # because the `basinmap0` feature is missing. Internally, dclab
        # locks up in a deep recursion involving ancillary features.
        with recursionlimit(100):  # speeds up this test
            with pytest.raises(BasinmapFeatureMissingError,
                               match="Could not find the feature 'basinmap0'"):
                dss["area_um"]


def test_export_to_hdf5_empty_features_mapped_basin():
    """Do not export any features, only the mapped basin information"""
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")
    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()
        ds.export.hdf5(
            path=path_exp,
            features=[],
            filtered=True,
            basins=True,
            )

    # Sanity check: Only the "deform" feature is in the dataset alongside
    # the basinmap feature, because this is a mapped basin.
    with h5py.File(path_exp) as h5:
        assert sorted(h5["events"].keys()) == ["basinmap0"]

    with dclab.new_dataset(path_exp) as dse:
        assert "deform" in dse
        assert "area_um" in dse
        assert np.allclose(dse["deform"][10], 0.057975024,
                           atol=1e-8, rtol=0)
        assert np.allclose(dse["area_um"][10], 79.0126,
                           atol=1e-3, rtol=0)
        assert "mapped" in str(dse.basins[0])


@pytest.mark.parametrize("filter1,filter2,offset", [
    # original dataset has "same" basins, filtering disabled
    [0, 0, 15],
    # original dataset has mapped basins, filtering disabled
    [10, 0, 5],
    # original dataset has "same" basins, filtering enabled
    [0, 10, 5],
    # original dataset has mapped basins, filtering enabled
    [5, 5, 5],
    [3, 7, 5],
])
def test_export_to_hdf5_inception(filter1, filter2, offset):
    """original dataset has "same" basins, filtering enabled"""
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp1 = h5path.with_name("exported.rtdc")
    path_exp2 = h5path.with_name("exported_inception.rtdc")

    with dclab.new_dataset(h5path) as ds:
        nevents = len(ds)
        ds.filter.manual[:filter1] = False
        ds.apply_filter()
        ds.export.hdf5(
            path=path_exp1,
            features=["deform"],
            filtered=bool(filter1),
            basins=True,
        )

    with dclab.new_dataset(path_exp1) as ds1:
        # sanity check
        assert len(ds1) == nevents - filter1
        ds1.filter.manual[:filter2] = False
        ds1.apply_filter()
        ds1.export.hdf5(
            path=path_exp2,
            features=["deform"],
            filtered=bool(filter2),
            basins=True,
        )

    # Test presence of basinmap features
    with h5py.File(path_exp2) as h5:
        feature_list = ["deform"]
        if filter1 or filter2:
            feature_list.append("basinmap0")
        if filter1 and filter2:
            feature_list.append("basinmap1")
        assert sorted(h5["events"].keys()) == sorted(feature_list)

    with dclab.new_dataset(path_exp2) as ds2:
        # sanity check
        assert len(ds2) == nevents - filter1 - filter2
        assert "deform" in ds2
        assert "area_um" in ds2
        assert np.allclose(ds2["deform"][offset], 0.057975024,
                           atol=1e-8, rtol=0)
        assert np.allclose(ds2["area_um"][offset], 79.0126,
                           atol=1e-3, rtol=0)
        # cross-check with the original dataset
        with dclab.new_dataset(h5path) as ds0:
            fsum = filter1 + filter2
            assert np.allclose(ds2["deform"], ds0["deform"][fsum:])
            assert np.allclose(ds2["area_um"], ds0["area_um"][fsum:])


def test_export_to_hdf5_mapped_basin():
    """When exporting filtered data to HDF5, a mapped basin is added"""
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    path_exp = h5path.with_name("exported.rtdc")
    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:5] = False
        ds.apply_filter()
        ds.export.hdf5(
            path=path_exp,
            features=["deform"],
            filtered=True,
            basins=True,
            )

    # Sanity check: Only the "deform" feature is in the dataset alongside
    # the basinmap feature, because this is a mapped basin.
    with h5py.File(path_exp) as h5:
        assert sorted(h5["events"].keys()) == ["basinmap0", "deform"]

    with dclab.new_dataset(path_exp) as dse:
        assert "deform" in dse
        assert "area_um" in dse
        assert np.allclose(dse["deform"][10], 0.057975024,
                           atol=1e-8, rtol=0)
        assert np.allclose(dse["area_um"][10], 79.0126,
                           atol=1e-3, rtol=0)
        assert "mapped" in str(dse.basins[0])
        # cross-check with the original dataset
        with dclab.new_dataset(h5path) as ds0:
            assert np.allclose(dse["deform"], ds0["deform"][5:])
            assert np.allclose(dse["area_um"], ds0["area_um"][5:])


@pytest.mark.skipif(not DCOR_AVAILABLE, reason="DCOR not available")
@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_export_to_hdf5_from_dcor_dataset(tmp_path):
    """Exporting from DCOR with mapped basin should make basin available"""
    pytest.importorskip("requests")
    # TODO: Implement this test with a new, smaller DCOR dataset
    dcor_url = ("https://dcor.mpl.mpg.de/api/3/action/dcserv?id="
                "fb719fb2-bd9f-817a-7d70-f4002af916f0")
    path_exp = tmp_path / "exported_from_dcor.rtdc"
    with dclab.new_dataset(dcor_url) as ds:
        ds.filter.manual[5:] = False
        ds.apply_filter()
        ds.export.hdf5(
            path=path_exp,
            features=["deform"],
            filtered=True,
            basins=True,
            )

    with dclab.new_dataset(path_exp) as dse:
        # exported feature
        assert "deform" in dse
        assert "mapped" in str(dse.basins[0])
        assert np.allclose(dse["deform"][0], 0.009741939,
                           atol=1e-8, rtol=0)
        # basin feature
        assert dse["image"][2][10, 10] == 54


def test_export_to_hdf5_no_mapped_basin_filters_disabled():
    """When exporting data without filters to HDF5, a normal basin is added"""
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    with h5py.File(h5path, "a") as h5:
        h5.attrs["experiment:run identifier"] = str(uuid.uuid4())

    path_exp = h5path.with_name("exported.rtdc")
    with dclab.new_dataset(h5path) as ds:
        ds.export.hdf5(
            path=path_exp,
            features=["deform"],
            filtered=False,
            basins=True,
            )

    # Sanity check: Only the "deform" feature is in the dataset.
    with h5py.File(path_exp) as h5:
        assert sorted(h5["events"].keys()) == ["deform"]

    with dclab.new_dataset(path_exp) as dse:
        assert "deform" in dse
        assert "area_um" in dse
        assert "mapped" not in str(dse.basins[0])
        assert np.allclose(dse["deform"][15], 0.057975024,
                           atol=1e-8, rtol=0)
        assert np.allclose(dse["area_um"][15], 79.0126,
                           atol=1e-3, rtol=0)

        # cross-check with the original dataset
        with dclab.new_dataset(h5path) as ds0:
            assert np.allclose(dse["deform"], ds0["deform"])
            assert np.allclose(dse["area_um"], ds0["area_um"])

    # And now a negative check with a modified basin identifier.
    with h5py.File(path_exp, "a") as h5e:
        h5e.attrs["experiment:run identifier"] = \
            h5e.attrs["experiment:run identifier"] + "-some-string"

    with dclab.new_dataset(path_exp) as dse2:
        assert "deform" in dse2
        assert "area_um" not in dse2  # sic
        assert np.allclose(dse2["deform"][15], 0.057975024,
                           atol=1e-8, rtol=0)


def test_sorting_priority_for_mapped_basins():
    """Basins are sorted by dclab

    Mapped basins should get lower priority.
    """
    mapper = "basinmap0"
    bnlist = [
        {"type": "remote", "format": "dcor", "mapping": "same", "ident": 0},
        {"type": "remote", "format": "dcor", "mapping": mapper, "ident": 1},
        {"type": "file", "format": "hdf5", "mapping": "same", "ident": 2},
        {"type": "file", "format": "hdf5", "mapping": mapper, "ident": 3},
        {"type": "hans", "format": "hdf5", "ident": 4},
        {"type": "remote", "format": "http", "ident": 5},
        {"type": "remote", "format": "http", "mapping": mapper, "ident": 6},

    ]
    sorted_list = sorted(bnlist, key=basin_priority_sorted_key)
    assert sorted_list[0]["ident"] == 2
    assert sorted_list[1]["ident"] == 3
    assert sorted_list[2]["ident"] == 5
    assert sorted_list[3]["ident"] == 6
    assert sorted_list[4]["ident"] == 0
    assert sorted_list[5]["ident"] == 1
    assert sorted_list[6]["ident"] == 4


def test_verify_basin_identifier():
    """The basin identifier for mapped basins must not match fully"""
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    # create a file-based basin with mapped content
    h5path_small = h5path.with_name("smaller.rtdc")
    mapping_array = np.array([1, 7, 10, 14], dtype=np.uint64)

    # Dataset creation
    with h5py.File(h5path, "a") as src, RTDCWriter(h5path_small) as hw:
        # write experiment identifiers
        hw.h5file.attrs.update(src.attrs)
        unique_id = f"unique_id_{uuid.uuid4()}"
        src.attrs["experiment:run identifier"] = unique_id
        hw.h5file.attrs["experiment:run identifier"] = unique_id + "-hans"

        assert src["events/deform"].size == 18, "sanity check"
        assert "area_um" in src["events"]

        # first, copy parts of the scalar features to the new file
        hw.store_feature("deform", src["events/deform"][mapping_array])
        # mapping array
        hw.store_feature("basinmap0", mapping_array)

        # Next, store the basin information in the new dataset
        bdat = {
            "type": "file",
            "format": "hdf5",
            "features": ["area_um"],
            "paths": [
                str(h5path),  # absolute path name
            ],
            "mapping": "basinmap0"
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = hw.h5file.require_group("basins")
        hw.write_text(basins, "invalid_format_basin", blines)

    with new_dataset(h5path) as ds0, new_dataset(h5path_small) as dsb:
        assert len(dsb.basins) == 1
        assert np.all(ds0["area_um"][mapping_array] == dsb["area_um"])
        assert np.all(ds0["deform"][mapping_array] == dsb["deform"])


def test_writer_reuse_basinmap_feature():
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    # create a file-based basin with mapped content
    h5path_small = h5path.with_name("smaller.rtdc")
    mapping_array = np.array([1, 7, 10, 14], dtype=np.uint64)

    with h5py.File(h5path, "a") as src, RTDCWriter(h5path_small) as hw:
        # write experiment identifiers
        hw.h5file.attrs.update(src.attrs)
        # store the basin once
        hw.store_basin(
            basin_name="exported with writer",
            basin_map=mapping_array,
            basin_locs=[h5path],
            basin_format="hdf5",
            basin_type="file",
            verify=False,
        )
        # store basin again, same mapping array
        hw.store_basin(
            basin_name="exported with writer",
            basin_map=mapping_array,
            basin_locs=["unknown.rtdc"],
            basin_format="hdf5",
            basin_type="file",
            verify=False,
        )

    # The essence of the test is to check that, even though two basins
    # were written, both use the same `mapping_array`.
    with h5py.File(h5path_small, "r") as h5:
        assert list(h5["events"].keys()) == ["basinmap0"]


def test_writer_verify_dataset_with_mapped_basin():
    h5path = retrieve_data("fmt-hdf5_image-mask-blood_2021.zip")
    # create a file-based basin with mapped content
    h5path_small = h5path.with_name("smaller.rtdc")
    mapping_array = np.array([1, 7, 10, 14], dtype=np.uint64)

    # Dataset creation
    with h5py.File(h5path, "a") as src, RTDCWriter(h5path_small) as hw:
        # write experiment identifiers
        hw.h5file.attrs.update(src.attrs)
        unique_id = f"unique_id_{uuid.uuid4()}"
        src.attrs["experiment:run identifier"] = unique_id
        hw.h5file.attrs["experiment:run identifier"] = unique_id + "-hans"

        assert src["events/deform"].size == 18, "sanity check"
        assert "area_um" in src["events"]

        # first, copy parts of the scalar features to the new file
        hw.store_feature("deform", src["events/deform"][mapping_array])

        # Write the mapped basin
        hw.store_basin(
            basin_name="exported with writer",
            basin_map=mapping_array,
            basin_locs=[h5path],
            basin_format="hdf5",
            basin_type="file",
            verify=True,  # this is the test
        )

    # Make sure this worked
    with dclab.new_dataset(h5path_small) as dse:
        assert "deform" in dse
        assert "area_um" in dse
        assert dse["image"][0][10, 10] == 73



"""Dataset overview plot

This example demonstrates basic data visualization with dclab and
matplotlib. To run this script, download the reference dataset
`calibration_beads.rtdc` :cite:`FigshareRef` and place it in the
same directory.

You will find more examples in the
:ref:`advanced usage <sec_advanced_scripting>` section of this
documentation.
"""
import matplotlib.pylab as plt
import numpy as np

import dclab

# Dataset to display
DATASET_PATH = "calibration_beads.rtdc"
# Features for scatter plot
SCATTER_X = "area_um"
SCATTER_Y = "deform"
# Event index to display
EVENT_INDEX = 100

xlabel = dclab.dfn.get_feature_label(SCATTER_X)
ylabel = dclab.dfn.get_feature_label(SCATTER_Y)

ds = dclab.new_dataset(DATASET_PATH)

fig = plt.figure(figsize=(8, 7))


ax1 = plt.subplot(221, title="Simple scatter plot")
ax1.plot(ds[SCATTER_X], ds[SCATTER_Y], "o", color="k", alpha=.2, ms=1)
ax1.set_xlabel(xlabel)
ax1.set_ylabel(ylabel)
ax1.set_xlim(19, 40)
ax1.set_ylim(0.005, 0.03)

ax2 = plt.subplot(222, title="KDE scatter plot")
sc = ax2.scatter(ds[SCATTER_X], ds[SCATTER_Y],
                 c=ds.get_kde_scatter(xax=SCATTER_X,
                                      yax=SCATTER_Y,
                                      kde_type="multivariate"),
                 s=3)
plt.colorbar(sc, label="kernel density [a.u]", ax=ax2)
ax2.set_xlabel(xlabel)
ax2.set_ylabel(ylabel)
ax2.set_xlim(19, 40)
ax2.set_ylim(0.005, 0.03)

ax3 = plt.subplot(425, title="Event image with contour")
ax3.imshow(ds["image"][EVENT_INDEX], cmap="gray")
ax3.plot(ds["contour"][EVENT_INDEX][:, 0],
         ds["contour"][EVENT_INDEX][:, 1],
         c="r")
ax3.set_xlabel("Detector X [px]")
ax3.set_ylabel("Detector Y [px]")

ax4 = plt.subplot(427, title="Event mask with µm-scale")
pxsize = ds.config["imaging"]["pixel size"]
ax4.imshow(ds["mask"][EVENT_INDEX],
           extent=[0, ds["mask"].shape[2] * pxsize,
                   0, ds["mask"].shape[1] * pxsize],
           cmap="gray")
ax4.set_xlabel("Detector X [µm]")
ax4.set_ylabel("Detector Y [µm]")

ax5 = plt.subplot(224, title="Fluorescence traces")
flsamples = ds.config["fluorescence"]["samples per event"]
flrate = ds.config["fluorescence"]["sample rate"]
fltime = np.arange(flsamples) / flrate * 1e6
# here we plot "fl?_raw"; you may also plot "fl?_med"
ax5.plot(fltime, ds["trace"]["fl1_raw"][EVENT_INDEX],
         c="#15BF00", label="fl1_raw")
ax5.plot(fltime, ds["trace"]["fl2_raw"][EVENT_INDEX],
         c="#BF8A00", label="fl2_raw")
ax5.plot(fltime, ds["trace"]["fl3_raw"][EVENT_INDEX],
         c="#BF0C00", label="fl3_raw")
ax5.legend()
ax5.set_xlim(ds["fl1_pos"][EVENT_INDEX] - 2*ds["fl1_width"][EVENT_INDEX],
             ds["fl1_pos"][EVENT_INDEX] + 2*ds["fl1_width"][EVENT_INDEX])
ax5.set_xlabel("Event time [µs]")
ax5.set_ylabel("Fluorescence [a.u.]")

plt.tight_layout()

plt.show()



"""Test HTTP format"""
import time

import numpy as np
import pytest


from dclab.rtdc_dataset.fmt_http import RTDC_HTTP

from helper_methods import DCOR_AVAILABLE


pytest.importorskip("requests")

if not DCOR_AVAILABLE:
    pytest.skip("No connection to DCOR", allow_module_level=True)


def test_cache_features():
    # This is the calibration beads measurement.
    # https://dcor.mpl.mpg.de/dataset/figshare-7771184-v2/
    # resource/fb719fb2-bd9f-817a-7d70-f4002af916f0
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")

    with RTDC_HTTP(s3_url) as ds:
        t0 = time.perf_counter()
        _ = ds["deform"][:]
        _ = ds["image"][10]
        t1 = time.perf_counter()
        for ii in range(50):
            _ = ds["deform"][:]
            _ = ds["image"][10]
        t2 = time.perf_counter()
        assert t2 - t1 < t1 - t0


def test_identifier():
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")

    with RTDC_HTTP(s3_url) as ds:
        # This is the HTTP ETag (https://en.wikipedia.org/wiki/HTTP_ETag)
        # given to this resource by the object store. If the file is
        # re-uploaded, the ETag may change and this test will fail.
        assert ds.identifier == "f0104b0ca2e7d6960189c60fc8b4b986-14"


@pytest.mark.parametrize("netloc", [
    "objectstore.hpccloud.mpcdf.mpg.de",
    "objectstore.hpccloud.mpcdf.mpg.de:443"
])
def test_netloc_vs_hostname(netloc):
    s3_url = (f"https://{netloc}/"
              f"circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              f"resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")
    with RTDC_HTTP(s3_url) as ds:
        assert len(ds) == 5000
        assert np.allclose(ds["deform"][0], 0.009741939,
                           rtol=0, atol=1e-7)


def test_open_public_s3_dataset():
    # This is the calibration beads measurement.
    # https://dcor.mpl.mpg.de/dataset/figshare-7771184-v2/
    # resource/fb719fb2-bd9f-817a-7d70-f4002af916f0
    s3_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
              "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
              "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")

    with RTDC_HTTP(s3_url) as ds:
        assert ds.config["experiment"]["sample"] == "calibration_beads"
        assert len(ds) == 5000
        assert np.allclose(ds["deform"][100], 0.013640802,
                           atol=0,
                           rtol=1e-5)



import json
import uuid

import h5py
import numpy as np

import pytest

import dclab
from dclab.http_utils import ETagNotInResponseHeaderWarning
from dclab import new_dataset, RTDCWriter
from dclab.rtdc_dataset.fmt_http import HTTPBasin, RTDC_HTTP
from dclab.rtdc_dataset.feat_basin import BasinNotAvailableError


from helper_methods import DCOR_AVAILABLE, retrieve_data


pytest.importorskip("requests")

if not DCOR_AVAILABLE:
    pytest.skip("No connection to DCOR", allow_module_level=True)


http_url = ("https://objectstore.hpccloud.mpcdf.mpg.de/"
            "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
            "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")


def test_basin_as_dict(tmp_path):
    tmp_path = tmp_path.resolve()
    h5path = tmp_path / "test_basin_http.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_HTTP(http_url) as src:
        # Store non-existent basin information
        with RTDCWriter(dst, mode="append") as hw:
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)
            hw.store_basin(basin_name="example basin",
                           basin_type="remote",
                           basin_format="http",
                           basin_locs=[http_url],
                           basin_descr="an example http test basin",
                           )

    with new_dataset(h5path) as ds:
        assert ds._enable_basins
        bdict = ds.basins[0].as_dict()
        assert bdict["basin_name"] == "example basin"
        assert bdict["basin_type"] == "remote"
        assert bdict["basin_format"] == "http"
        assert bdict["basin_locs"] == [http_url]
        assert bdict["basin_descr"] == "an example http test basin"

    # Now use the data from `bdict` to create a new basin
    h5path_two = h5path.with_name("smaller_two.rtdc")

    # Dataset creation
    with RTDCWriter(h5path_two) as hw:
        # first, copy all the scalar features to the new file
        hw.store_metadata(meta)
        hw.store_basin(**bdict)

    with new_dataset(h5path_two) as ds2:
        bdict2 = ds2.basins[0].as_dict()
        assert bdict2["basin_name"] == "example basin"
        assert bdict2["basin_type"] == "remote"
        assert bdict2["basin_format"] == "http"
        assert bdict2["basin_locs"] == [http_url]
        assert bdict2["basin_descr"] == "an example http test basin"


def test_basin_as_dict_netloc_vs_hostname(tmp_path):
    tmp_path = tmp_path.resolve()
    h5path = tmp_path / "test_basin_http.rtdc"

    # note the port is included here
    http_url_netloc = ("https://objectstore.hpccloud.mpcdf.mpg.de:443/"
                       "circle-5a7a053d-55fb-4f99-960c-f478d0bd418f/"
                       "resource/fb7/19f/b2-bd9f-817a-7d70-f4002af916f0")

    with h5py.File(h5path, "a") as dst, RTDC_HTTP(http_url) as src:
        # Store non-existent basin information
        with RTDCWriter(dst, mode="append") as hw:
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)
            hw.store_basin(basin_name="example basin",
                           basin_type="remote",
                           basin_format="http",
                           basin_locs=[http_url_netloc],
                           basin_descr="an example http test basin",
                           )

    with new_dataset(h5path) as ds:
        assert len(ds) == 5000
        # This failed in <0.56.0, because `netloc` was used instead of
        # `hostname` when connecting to the socket to check whether the
        # server is available.
        assert np.allclose(ds["deform"][0], 0.009741939,
                           rtol=0, atol=1e-7)


@pytest.mark.parametrize("url", [
    "https://example.com/nonexistentbucket/nonexistentkey",
    f"https://objectstore.hpccloud.mpcdf.mpg.de/noexist-{uuid.uuid4()}/key",
])
def test_basin_not_available(url):
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")

    # Dataset creation
    with h5py.File(h5path, "a") as dst:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "http",
            "urls": [
                # does not exist
                url
            ]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)

    # Open the dataset and check whether basin is missing
    with new_dataset(h5path) as ds:
        assert not ds.features_basin
        # This is a very subtle test for checking whether invalid basins
        # are just ignored:
        _ = ds["index"]

    # Also test that on a lower level
    bn = HTTPBasin("https://example.com/nonexistentbucket/nonexistentkey")
    assert not bn.is_available()
    with pytest.raises(BasinNotAvailableError, match="is not available"):
        _ = bn.ds


def test_create_basin_file_non_matching_identifier(tmp_path):
    h5path = tmp_path / "test_basin_http.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_HTTP(http_url) as src:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "http",
            "urls": [http_url],
            "features": ["deform"],
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
            meta = src.config.as_dict(pop_filtering=True)
            meta["experiment"]["run identifier"] = "hoolahoop"
            hw.store_metadata(meta)

    with new_dataset(h5path) as ds:
        assert ds.basins
        # The feature shows up as available...
        assert ds.features_basin == ["deform"]
        # ...but it is actually not, since the run identifier does not match
        # and therefore dclab does not allow the user to access it.
        with (pytest.warns(UserWarning, match="but I cannot get its data"),
              pytest.raises(KeyError, match="deform")):
            _ = ds["deform"]


def test_create_basin_file_with_no_data(tmp_path):
    h5path = tmp_path / "test_basin_http.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_HTTP(http_url) as src:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "http",
            "urls": [http_url]
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)

    with new_dataset(h5path) as ds:
        assert ds.features_basin
        assert len(ds) == 5000
        assert np.allclose(ds["deform"][0], 0.009741939,
                           atol=0, rtol=1e-5)


def test_create_basin_file_with_one_feature(tmp_path):
    h5path = tmp_path / "test_basin_http.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_HTTP(http_url) as src:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "http",
            "urls": [http_url],
            "features": ["deform"],
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)

    with new_dataset(h5path) as ds:
        assert ds.features_basin
        assert len(ds) == 5000
        assert "deform" in ds.features_basin
        assert "area_um" not in ds.features_basin
        assert "deform" in ds
        assert "area_um" not in ds
        assert np.allclose(ds["deform"][0], 0.009741939,
                           atol=0, rtol=1e-5)


def test_trace_availability(tmp_path):
    h5path = tmp_path / "test_basin_http.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_HTTP(http_url) as src:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "http",
            "urls": [http_url],
            "features": ["trace"],
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[0] = False
        with dclab.new_dataset(ds) as ds2:
            assert "trace" in ds2


def test_trace_availability_invalid(tmp_path):
    h5path = tmp_path / "test_basin_http.rtdc"

    with h5py.File(h5path, "a") as dst, RTDC_HTTP(http_url) as src:
        # Store non-existent basin information
        bdat = {
            "type": "remote",
            "format": "http",
            "urls": ["https://dcor.mpl.mpg.de/api/3/action/site_read"],
            "features": ["trace"],
        }
        blines = json.dumps(bdat, indent=2).split("\n")
        basins = dst.require_group("basins")
        with RTDCWriter(dst, mode="append") as hw:
            hw.write_text(basins, "my_basin", blines)
            meta = src.config.as_dict(pop_filtering=True)
            hw.store_metadata(meta)

    with dclab.new_dataset(h5path) as ds:
        ds.filter.manual[:] = False
        ds.filter.manual[:2] = True
        ds.apply_filter()
        assert "trace" in ds
        # Until a workaround is found for invalid basin URLs that return a
        # status code of 200, do this test which should raise a warning,
        # because `__contains__` returns True for "trace", but the trace data
        # are nowhere to find.
        with (pytest.warns(UserWarning, match="but I cannot get its data"),
              pytest.warns(ETagNotInResponseHeaderWarning,
                           match="Got empty ETag header"),
              pytest.raises(KeyError, match="trace")):
            _ = ds["trace"]
        with (pytest.warns(UserWarning, match="but I cannot get its data"),
              pytest.warns(ETagNotInResponseHeaderWarning,
                           match="Got empty ETag header"),
              pytest.raises(KeyError, match="trace")):
            dclab.new_dataset(ds)



import numbers
import re
from inspect import getmembers, isfunction

import numpy as np

import pytest

from dclab.definitions import meta_logic, meta_parse


def assert_is_bool_false(value):
    assert isinstance(value, bool)
    assert not value


def assert_is_bool_true(value):
    assert isinstance(value, bool)
    assert value


@pytest.mark.parametrize("exists,sec,key", [
    [True, "user", "test"],
    [False, "user", 1],
    [True, "setup", "channel width"],
    [False, "setup", "funnel width"],
    [True, "online_filter", "area_um,deform soft limit"],
    [False, "online_filter", "peter,deform soft limit"],
    [True, "online_filter", "area_um,deform polygon points"],
    [False, "online_filter", "peter,deform polygon points"],
    [True, "online_filter", "pos_x max"],
    [True, "online_filter", "pos_x min"],
    [True, "online_filter", "pos_x soft limit"],
    [True, "online_filter", "fl1_max max"],
    [True, "online_filter", "fl2_max min"],
    [True, "online_filter", "fl3_max soft limit"],
    [True, "online_filter", "target duration"],
    [True, "online_filter", "area_ratio max"],
])
def test_meta_logic_config_key_exists(exists, sec, key):
    if exists:
        assert meta_logic.config_key_exists(sec, key)
    else:
        assert not meta_logic.config_key_exists(sec, key)


@pytest.mark.parametrize("sec,key,descr", [
    ["user", "key", "key"],
    ["experiment", "date", "Date of measurement ('YYYY-MM-DD')"],
    ["online_filter",
     "area_um,deform soft limit",
     "Soft limit, polygon (Area, Deformation)"],
    ["online_filter",
     "area_um,deform polygon points",
     "Polygon (Area, Deformation)"],
    ["online_filter", "fl1_max max", "Max. FL-1 maximum [a.u.]"],
    ["online_filter", "fl1_max min", "Min. FL-1 maximum [a.u.]"],
    ["online_filter",
     "area_ratio soft limit",
     "Soft limit, Porosity (convex to measured area ratio)"],
    ["online_filter", "fl3_max soft limit", "Soft limit, FL-3 maximum"],
    ["online_filter", "target duration", "Target measurement duration [min]"],
    ["online_filter", "pos_x max", "Max. Position along channel axis [µm]"],
])
def test_meta_logic_get_config_value_descr(sec, key, descr):
    assert meta_logic.get_config_value_descr(sec, key) == descr


@pytest.mark.parametrize("sec,key,func", [
    ["experiment", "date", str],
    ["online_filter", "area_um,deform soft limit", meta_parse.fbool],
    ["online_filter",
     "area_um,deform polygon points",
     meta_parse.f2dfloatarray],
])
def test_meta_logic_get_config_value_func(sec, key, func):
    assert meta_logic.get_config_value_func(sec, key) is func


def test_meta_logic_get_config_value_func_user():
    lamb = meta_logic.get_config_value_func("user", "key")
    assert lamb("peter") == "peter"


@pytest.mark.parametrize("sec,key,dtype", [
    ["user", "key", None],
    ["experiment", "date", str],
    ["online_filter", "area_um,deform soft limit", bool],
    ["online_filter",
     "area_um,deform polygon points",
     np.ndarray],
    ["online_filter", "fl1_max max", numbers.Number],
    ["online_filter", "pos_x min", numbers.Number],
])
def test_meta_logic_get_config_value_type(sec, key, dtype):
    this_type = meta_logic.get_config_value_type(sec, key)
    if isinstance(this_type, (tuple, list)):
        assert dtype in this_type
    else:
        assert dtype is this_type


def test_meta_parse_f2dfloatarray():
    assert meta_parse.f2dfloatarray([[1], [2]]).shape == (2, 1)


def test_meta_parse_f1dfloatduple():
    assert meta_parse.f1dfloatduple([1, 2]) == (1.0, 2.0)
    assert meta_parse.f1dfloatduple((8, 9)) == (8.0, 9.0)
    assert meta_parse.f1dfloatduple([-8.999, 0.332]) == (-8.999, 0.332)
    assert meta_parse.f1dfloatduple((-8.999, 0.332)) == (-8.999, 0.332)
    assert meta_parse.f1dfloatduple(np.array([3, -1])) == (3.0, -1.0)
    with pytest.raises(ValueError, match="Value must be of length two, "
                                         "got length 3!"):
        # the given value must be 1d
        assert meta_parse.f1dfloatduple((3, -1, 5.6))
    with pytest.raises(ValueError, match=re.escape(
            "Value is not 1 dimensional, "
            "got [[ 3.  -1. ]\n [ 4.   5.5]]!")):
        # the given value must be 1d
        assert meta_parse.f1dfloatduple(np.array([[3, -1], [4, 5.5]]))

    # check the func_types mapping
    typ = meta_parse.func_types[meta_parse.f1dfloatduple]
    assert isinstance((1.0, 2.0), typ)


def test_meta_parse_fbool():
    assert_is_bool_true(meta_parse.fbool(True))
    assert_is_bool_true(meta_parse.fbool("true"))
    assert_is_bool_true(meta_parse.fbool("True"))
    assert_is_bool_true(meta_parse.fbool("1"))
    assert_is_bool_true(meta_parse.fbool(1))
    assert_is_bool_true(meta_parse.fbool(2.2))

    assert_is_bool_false(meta_parse.fbool(False))
    assert_is_bool_false(meta_parse.fbool("false"))
    assert_is_bool_false(meta_parse.fbool("False"))
    assert_is_bool_false(meta_parse.fbool("0"))
    assert_is_bool_false(meta_parse.fbool("0.0"))
    assert_is_bool_false(meta_parse.fbool(0))
    assert_is_bool_false(meta_parse.fbool(0.))

    with pytest.raises(ValueError, match="Empty string provided for fbool!"):
        meta_parse.fbool("")


def test_meta_parse_fboolorfloat():
    """Check fboolorfloat format"""
    assert meta_parse.fboolorfloat(True) is True
    assert meta_parse.fboolorfloat(False) is False
    assert meta_parse.fboolorfloat("true") is True
    assert meta_parse.fboolorfloat("False") is False

    assert meta_parse.fboolorfloat(2.0) == 2.0
    assert meta_parse.fboolorfloat(-0.44) == -0.44

    assert meta_parse.fboolorfloat(1) == 1.0
    assert meta_parse.fboolorfloat(-42) == -42.0

    with pytest.raises(ValueError, match=re.escape(
            "Value could not be converted to bool or float, got [1, 2, 3]!")):
        assert meta_parse.fboolorfloat([1, 2, 3])

    # check the func_types mapping
    typ = meta_parse.func_types[meta_parse.fboolorfloat]
    assert isinstance(True, typ)
    assert isinstance(2.0, typ)


def test_meta_parse_fint():
    assert meta_parse.fint(True) == 1
    assert meta_parse.fint("true") == 1
    assert meta_parse.fint("True") == 1
    assert meta_parse.fint(1) == 1
    assert meta_parse.fint(1.0) == 1
    assert meta_parse.fint("1") == 1

    assert meta_parse.fint(False) == 0
    assert meta_parse.fint("False") == 0
    assert meta_parse.fint("false") == 0
    assert meta_parse.fint("0") == 0
    assert meta_parse.fint("0.0") == 0
    assert meta_parse.fint(0) == 0
    assert meta_parse.fint(0.0) == 0

    assert meta_parse.fint("20") == 20
    assert meta_parse.fint(20.) == 20

    with pytest.raises(ValueError, match="Empty string provided for fint!"):
        meta_parse.fint("")


def test_meta_parse_fintlist():
    assert meta_parse.fintlist([True, 1, "1", 20.]) == [1, 1, 1, 20]
    assert meta_parse.fintlist("[True, 1, 1, 20.]") == [1, 1, 1, 20]


def test_meta_parse_lcstr():
    assert meta_parse.lcstr("PETER") == "peter"
    assert meta_parse.lcstr("Hans") == "hans"


def test_meta_parse_function_mapping():
    """Each function should have type(s) mapping in `meta_parse.func_types`"""

    meta_parse_funcs = getmembers(meta_parse, isfunction)
    type_mapping = list(meta_parse.func_types.keys())

    for meta_parse_func in meta_parse_funcs:
        assert meta_parse_func[1] in type_mapping, (
            f"The `meta_parse` module function `{meta_parse_func}` does not "
            "have a type mapping in meta_parse.func_types.")



"""Extract emodulus look-up table (LUT) and isoelastics from simulation data.

See lut_processor for technical details.

Note that a matplotlib window will open when the isoelastics are
created, so you can verify that everything is in order. Just close
that window to proceed.
"""
import argparse
import copy
import json
import pathlib

import matplotlib.pylab as plt
import numpy as np

from lut_recipes import LutProcessor


def main():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('input', metavar="INPUT", type=str,
                        help='Input path (.hdf5 file)')
    parser.add_argument("--raw",
                        dest='raw',
                        action='store_true',
                        help="do not perform data post-processing",
                        )
    parser.set_defaults(raw=False)

    args = parser.parse_args()
    path = pathlib.Path(args.input)
    raw = args.raw

    if raw:
        print("Skipping all post-processing steps!")

    print("Extracting LUT and isoelastics")
    lup = LutProcessor(path, use_hooks=not raw, verbose=True)
    lut, contours, contour_levels = lup.assemble_lut_and_isoelastics()

    ax = plt.subplot(111, title="Final LUT and isoelastics")
    ax.plot(lut[:, 0], lut[:, 1], ".", color="k")
    for cc in contours:
        ax.plot(cc[:, 0], cc[:, 1])
    plt.show()

    print("Saving LUT and isoelastics")
    save_iso(path.with_name(path.name.rsplit(".", 1)[0] + "_iso.txt"),
             contours, contour_levels, lup.meta)

    save_lut(path.with_name(path.name.rsplit(
        ".", 1)[0] + "_lut.txt"), lut, lup.meta)


def save_iso(path, contours, levels, meta, header=None):
    """Save isoelastics to a text file for usage in dclab"""
    # change identifier to lut identifier
    if header is None:
        header = ["area_um [um^2]", "deform", "emodulus [kPa]"]
    meta = copy.deepcopy(meta)
    if "identifier" in meta:
        idx = meta.pop("identifier")
        meta["lut identifier"] = idx

    lentot = 0
    for cc in contours:
        lentot += len(cc)

    cdat = np.zeros((lentot, 3), dtype=float)
    ii = 0
    for cc, lev in zip(contours, levels):
        sli = slice(ii, ii + len(cc))
        cdat[sli, 0] = cc[:, 0]
        cdat[sli, 1] = cc[:, 1]
        cdat[sli, 2] = lev
        ii += len(cc)

    assert ii == lentot

    save_lut(path, cdat, meta, header=header)


def save_lut(path, lut, meta, header=None):
    """Save LUT to a text file for usage in dclab"""
    if header is None:
        header = ["area_um [um^2]", "deform", "emodulus [kPa]"]
    meta = copy.deepcopy(meta)
    if "column features" in meta:
        # sanity check
        for ii, kk in enumerate(meta["column features"]):
            assert kk in header[ii]
        # remove redundant information; this is stored in the header
        meta.pop("column features")
    with path.open("w") as fd:
        # Metadata
        dump = json.dumps(meta, sort_keys=True, indent=2)
        fd.write("# BEGIN METADATA\r\n")
        for dl in dump.split("\n"):
            fd.write("# " + dl + "\r\n")
        fd.write("# END METADATA\r\n")
        fd.write("#\r\n")

        # Header
        fd.write("# " + "\t".join(header) + "\r\n")

        # Data
        for ii in range(lut.shape[0]):
            line = "{:.5e}\t{:.5e}\t{:.5e}\r\n".format(*lut[ii])
            fd.write(line)


if __name__ == "__main__":
    main()



import h5py
import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.features import inert_ratio as ir

from helper_methods import retrieve_data


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_af_inert_ratio_cvx():
    pytest.importorskip("nptdms")
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    # This is something low-level and should not be done in a script.
    # Remove the brightness columns from RTDCBase to force computation with
    # the image and contour columns.
    real_ir = ds._events.pop("inert_ratio_cvx")
    # This will cause a zero-padding warning:
    comp_ir = ds["inert_ratio_cvx"]
    idcompare = ~np.isnan(comp_ir)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_ir[idcompare], comp_ir[idcompare])


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_af_inert_ratio_prnc():
    pytest.importorskip("nptdms")
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    # This will cause a zero-padding warning:
    prnc = ds["inert_ratio_prnc"]
    raw = ds["inert_ratio_raw"]
    idcompare = ~np.isnan(prnc)
    # ignore first event (no image data)
    idcompare[0] = False
    diff = (prnc - raw)[idcompare]
    # only compare the first valid event which seems to be quite close
    assert np.allclose(diff[0], 0, atol=1.2e-3, rtol=0)


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
@pytest.mark.filterwarnings(
    'ignore::dclab.rtdc_dataset.config.UnknownConfigurationKeyWarning')
@pytest.mark.parametrize("file", [
    "fmt-hdf5_fl_2017.zip",
    "fmt-hdf5_mask-contour_2018.zip",
    "fmt-hdf5_image-bg_2020.zip",
    "fmt-hdf5_polygon_gate_2021.zip",
    "fmt-hdf5_fl_wide-channel_2023.zip",
    ])
def test_af_inert_ratio_prnc_greater_than_one_issue_212(file):
    ds = new_dataset(retrieve_data(file))
    prnc = ds["inert_ratio_prnc"]
    assert np.all(prnc >= 1)


@pytest.mark.filterwarnings('ignore::dclab.rtdc_dataset.'
                            + 'feat_anc_core.ancillary_feature.'
                            + 'BadFeatureSizeWarning')
def test_af_inert_ratio_raw():
    pytest.importorskip("nptdms")
    # Brightness of the image
    ds = dclab.new_dataset(retrieve_data("fmt-tdms_fl-image-bright_2017.zip"))
    # This is something low-level and should not be done in a script.
    # Remove the brightness columns from RTDCBase to force computation with
    # the image and contour columns.
    real_ir = ds._events.pop("inert_ratio_raw")
    # This will cause a zero-padding warning:
    comp_ir = ds["inert_ratio_raw"]
    idcompare = ~np.isnan(comp_ir)
    # ignore first event (no image data)
    idcompare[0] = False
    assert np.allclose(real_ir[idcompare], comp_ir[idcompare])


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_inert_ratio_raw():
    ds = new_dataset(retrieve_data("fmt-hdf5_mask-contour_2018.zip"))

    raw = ir.get_inert_ratio_raw(cont=ds["contour"])
    ref = np.array([4.25854232,  1.22342663,  4.64971179,  1.70914857,
                    3.62797492, 1.51502192,  2.74757573,  1.79841136])
    assert np.allclose(ref, raw, rtol=0, atol=5e-9)


def test_inert_ratio_raw_correct_issue_224():
    path = retrieve_data("fmt-hdf5_wide-channel_2023.zip")

    # make the file look like it has been created with an old Shape-In version
    with h5py.File(path, "a") as h5:
        h5["events/inert_ratio_cvx"][:] = 1
        assert np.allclose(h5["events/inert_ratio_cvx"][:], 1)
        h5.attrs["setup:software version"] = "ShapeIn 2.0.4 | dclab 0.49.0"

    ds = dclab.new_dataset(path)
    assert "inert_ratio_cvx" in ds.features_loaded
    assert np.allclose(ds["inert_ratio_cvx"], 1)


def test_inert_ratio_raw_with_integer_precision_issue_223():
    """Test computation of inert_ratio_raw

    On Windows, we had the problem that due to the default integer
    not being int64, the inertia ratio was not computed correctly.
    """
    path = retrieve_data("fmt-hdf5_wide-channel_2023.zip")

    # make the file look like it has been created with an old Shape-In version
    with h5py.File(path, "a") as h5:
        inert_ratio_raw = h5["events/inert_ratio_raw"][:]
        del h5["events/inert_ratio_raw"]

    ds = dclab.new_dataset(path)
    assert "inert_ratio_raw" not in ds.features_loaded
    assert np.allclose(inert_ratio_raw,
                       ds["inert_ratio_raw"],
                       atol=0,
                       rtol=1e-5,
                       equal_nan=False,  # sic!
                       )
    assert not np.sum(np.isnan(ds["inert_ratio_raw"]))


def test_inert_ratio_raw_with_old_dclab_issue_224():
    path = retrieve_data("fmt-hdf5_wide-channel_2023.zip")

    # make the file look like it has been created with an old Shape-In version
    with h5py.File(path, "a") as h5:
        h5.attrs["setup:software version"] = "ShapeIn 2.0.4 | dclab 0.48.0"

    ds = dclab.new_dataset(path)
    assert "inert_ratio_cvx" not in ds.features_loaded


def test_inert_ratio_raw_with_old_shapein_issue_224():
    path = retrieve_data("fmt-hdf5_wide-channel_2023.zip")

    # make the file look like it has been created with an old Shape-In version
    with h5py.File(path, "a") as h5:
        h5["events/inert_ratio_cvx"][:] = 1
        assert np.allclose(h5["events/inert_ratio_cvx"][:], 1)
        h5.attrs["setup:software version"] = "ShapeIn 2.0.4"

    ds = dclab.new_dataset(path)
    assert "inert_ratio_cvx" in ds.features_loaded


@pytest.mark.parametrize("si_string", ["ShapeIn 2.0.5", "2.3.1"])
def test_inert_ratio_with_new_shapein_issue_224(si_string):
    path = retrieve_data("fmt-hdf5_wide-channel_2023.zip")

    # make the file look like it has been created with an old Shape-In version
    with h5py.File(path, "a") as h5:
        h5["events/inert_ratio_cvx"][:] = 1
        assert np.allclose(h5["events/inert_ratio_cvx"][:], 1)
        h5.attrs["setup:software version"] = si_string + " | dclab 0.47.2"
        # Create a Shape-In acquisition log so that it is not considered a
        # defective feature.
        h5["logs/shapein-acquisition"] = ["test"]

    ds = dclab.new_dataset(path)
    assert "inert_ratio_cvx" in ds.features_loaded


def test_inert_ratio_prnc():
    """Test equivalence of inert_ratio_raw and inert_ratio_prnc"""
    t = np.linspace(0, 2*np.pi, 300)

    x1 = 1.7 * np.cos(t)
    y1 = 1.1 * np.sin(t)
    c1 = np.dstack((x1, y1))[0]
    raw = ir.get_inert_ratio_raw(c1)

    phi = np.arctan2(y1, x1)
    rho = np.sqrt(x1**2 + y1**2)

    for theta in np.linspace(0.1, 2*np.pi, 14):  # arbitrary rotation
        for pos_x in np.linspace(-5, 20, 8):  # arbitrary x shift
            for pos_y in np.linspace(-4.6, 17, 4):  # arbitrary y shift
                x2 = rho * np.cos(phi + theta) + pos_x
                y2 = rho * np.sin(phi + theta) + pos_y

                c2 = np.dstack((x2, y2))[0]
                prnc = ir.get_inert_ratio_prnc(c2)

                assert np.allclose(raw, prnc, rtol=0, atol=1e-7)


def test_inert_ratio_prnc_simple_1():
    c = np.array([[0, 0],
                  [0, 1],
                  [0, 2],
                  [1, 2],
                  [2, 2],
                  [3, 2],
                  [3, 1],
                  [3, 0],
                  [2, 0],
                  [1, 0],
                  [0, 0]])
    raw = ir.get_inert_ratio_raw(c)
    prnc = ir.get_inert_ratio_prnc(c)
    tilt = ir.get_tilt(c)
    assert np.allclose(raw, 1.5)
    assert np.allclose(prnc, 1.5)
    assert np.allclose(tilt, 0)


def test_inert_ratio_prnc_simple_2():
    c = np.array([[0, 0],
                  [1, 1],
                  [2, 2],
                  [3, 3],
                  [4, 2],
                  [5, 1],
                  [4, 0],
                  [3, -1],
                  [2, -2],
                  [1, -1],
                  [0, 0]])
    raw = ir.get_inert_ratio_raw(c)
    prnc = ir.get_inert_ratio_prnc(c)
    tilt = ir.get_tilt(c)
    assert np.allclose(raw, 1)
    assert np.allclose(prnc, 1.5)
    assert np.allclose(tilt, np.pi/4)


def test_tilt():
    t = np.linspace(0, 2*np.pi, 300)

    x1 = 1.7 * np.cos(t)
    y1 = 1.1 * np.sin(t)

    phi = np.arctan2(y1, x1)
    rho = np.sqrt(x1**2 + y1**2)

    for theta in np.linspace(-.3, 2.2*np.pi, 89):  # arbitrary rotation
        x2 = rho * np.cos(phi + theta)
        y2 = rho * np.sin(phi + theta)

        c2 = np.dstack((x2, y2))[0]
        tilt = ir.get_tilt(c2)

        th = np.mod(theta, np.pi)
        if th > np.pi/2:
            th -= np.pi
        th = np.abs(th)
        assert np.allclose(tilt, th)



import hdf5plugin
import h5py
import numpy as np
import pytest

import dclab
from dclab import new_dataset
from dclab.rtdc_dataset import is_properly_compressed, rtdc_copy
from dclab.rtdc_dataset import RTDCWriter

from helper_methods import retrieve_data


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_copy_already_compressed():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_temp = path.with_name("test_compressed.rtdc")
    path_copy = path.with_name("test_copy.rtdc")

    ds1 = new_dataset(path)

    with RTDCWriter(path_temp,
                    compression_kwargs=hdf5plugin.Zstd(clevel=5)) as hw:
        hw.store_metadata({"setup": ds1.config["setup"],
                           "experiment": ds1.config["experiment"]})
        hw.store_feature("deform", ds1["deform"])
        hw.store_feature("image", ds1["image"])

        # sanity check
        assert is_properly_compressed(hw.h5file["events/deform"])
        assert is_properly_compressed(hw.h5file["events/image"])

        with h5py.File(path_copy, "w") as hc:
            rtdc_copy(src_h5file=hw.h5file,
                      dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["events/deform"])
        assert is_properly_compressed(hc["events/image"])


def test_copy_basins():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # add log to source file
    with RTDCWriter(path, mode="append") as hw:
        bn_hash = hw.store_basin(
            basin_type="file",
            basin_format="hdf5",
            basin_name="test_basin",
            basin_locs=["does-not-exist-but-does-not-matter.rtdc"],
            verify=False
        )
        assert not is_properly_compressed(hw.h5file[f"basins/{bn_hash}"])

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["events/deform"])
        assert is_properly_compressed(hc[f"basins/{bn_hash}"])


def test_copy_basins_internal_none():
    """Store an internal basin and only export scalar features"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("input_file.rtdc")
    h5path_out = h5path.with_name("condensed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_copy(src_h5file=src,
                  dst_h5file=hw.h5file,
                  features="scalar")
        assert "basins" not in hw.h5file, "no basins in input file"
        # store scalar and non-scalar internal basins in the input file
        hw.store_basin(basin_name="scalar and non-scalar basin data",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data={"image_bg": np.zeros((2, 80, 320))},
                       basin_map=np.zeros(src["events/deform"].shape[0]),
                       basin_feats=["image_bg"],
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "image_bg" in ds.features_basin
        assert "image_bg" not in ds.features_innate

    # Now the actual tests starts.
    with h5py.File(h5path_small) as src, h5py.File(h5path_out, "a") as dst:
        rtdc_copy(src_h5file=src,
                  dst_h5file=dst,
                  features="scalar",
                  include_basins=True,
                  include_logs=True,
                  include_tables=True,
                  meta_prefix="")

    with h5py.File(h5path_out) as h5:
        # The output file should not contain the "image_bg" data.
        assert "basin_events" not in h5

    # When we load the basin data, "image_bg" should also not be there.
    with dclab.new_dataset(h5path_out) as ds:
        basin_dicts = ds.basins_get_dicts()
        assert len(basin_dicts) == 0


def test_copy_basins_internal_no_scalar():
    """Store an internal basin and only export scalar features"""
    h5path = retrieve_data("fmt-hdf5_fl_wide-channel_2023.zip")
    h5path_small = h5path.with_name("input_file.rtdc")
    h5path_out = h5path.with_name("condensed.rtdc")

    # Dataset creation
    with h5py.File(h5path) as src, RTDCWriter(h5path_small) as hw:
        # first, copy all the scalar features to the new file
        rtdc_copy(src_h5file=src,
                  dst_h5file=hw.h5file,
                  features="scalar")
        assert "basins" not in hw.h5file, "no basins in input file"
        # store scalar and non-scalar internal basins in the input file
        hw.store_basin(basin_name="scalar and non-scalar basin data",
                       basin_type="internal",
                       basin_format="h5dataset",
                       basin_locs=["basin_events"],
                       basin_descr="an example test basin",
                       internal_data={"userdef1": np.arange(2),
                                      "image_bg": np.zeros((2, 80, 320)),
                                      },
                       basin_map=np.zeros(src["events/deform"].shape[0]),
                       basin_feats=["image_bg", "userdef1"],
                       )

    # sanity check
    with new_dataset(h5path_small) as ds:
        assert "userdef1" in ds.features_basin
        assert "userdef1" not in ds.features_innate
        assert "image_bg" in ds.features_basin
        assert "image_bg" not in ds.features_innate

    # Now the actual tests starts.
    with h5py.File(h5path_small) as src, h5py.File(h5path_out, "a") as dst:
        rtdc_copy(src_h5file=src,
                  dst_h5file=dst,
                  features="scalar",
                  include_basins=True,
                  include_logs=True,
                  include_tables=True,
                  meta_prefix="")

    with h5py.File(h5path_out) as h5:
        # The output file should not contain the "image_bg" data.
        assert "userdef1" in h5["basin_events"]
        assert "image_bg" not in h5["basin_events"]

    # When we load the basin data, "image_bg" should also not be there.
    with dclab.new_dataset(h5path_out) as ds:
        basin_dicts = ds.basins_get_dicts()
        assert len(basin_dicts) == 1
        bn = basin_dicts[0]
        assert bn["type"] == "internal"
        assert bn["mapping"] != "same"
        assert bn["features"] == ["userdef1"], "image_bg feature must be gone"


def test_copy_basins_mapped():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # add log to source file
    with RTDCWriter(path, mode="append") as hw:
        bn_hash = hw.store_basin(
            basin_type="file",
            basin_format="hdf5",
            basin_name="test_basin",
            basin_locs=["does-not-exist-but-does-not-matter.rtdc"],
            basin_map=np.arange(len(hw.h5file["events/deform"])),
            verify=False
        )
        assert not is_properly_compressed(hw.h5file[f"basins/{bn_hash}"])
        assert not is_properly_compressed(hw.h5file["events/basinmap0"])

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["events/deform"])
        assert is_properly_compressed(hc[f"basins/{bn_hash}"])
        assert is_properly_compressed(hc["events/basinmap0"])


def test_copy_basins_multiple():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # add log to source file
    with RTDCWriter(path, mode="append") as hw:
        hw.store_basin(
            basin_type="file",
            basin_format="hdf5",
            basin_name="test_basin",
            basin_locs=["does-not-exist-but-does-not-matter.rtdc"],
            basin_map=np.arange(len(hw.h5file["events/deform"])),
            basin_descr="A first basin",
            verify=False
        )

        hw.store_basin(
            basin_type="file",
            basin_format="hdf5",
            basin_name="test_basin",
            basin_locs=["does-not-exist-but-does-not-matter.rtdc"],
            basin_map=np.arange(len(hw.h5file["events/deform"])),
            basin_descr="A second basin",
            verify=False
        )

        assert len(hw.h5file["basins"]) == 2

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert len(hc["basins"]) == 2


def test_copy_basins_no_basin():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # add log to source file
    with RTDCWriter(path, mode="append") as hw:
        bn_hash = hw.store_basin(
            basin_type="file",
            basin_format="hdf5",
            basin_name="test_basin",
            basin_locs=["does-not-exist-but-does-not-matter.rtdc"],
            basin_map=np.arange(len(hw.h5file["events/deform"])),
            verify=False
        )
        assert not is_properly_compressed(hw.h5file[f"basins/{bn_hash}"])
        assert not is_properly_compressed(hw.h5file["events/basinmap0"])

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc,
                  include_basins=False)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["events/deform"])
        assert "basins" not in hc
        assert "basinmap0" not in hc["events"]


def test_copy_logs():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # add log to source file
    with RTDCWriter(path, mode="append") as hw:
        hw.store_log("test_log", ["hans", "peter", "und", "zapadust"])
        assert not is_properly_compressed(hw.h5file["logs/test_log"])

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["events/deform"])
        assert is_properly_compressed(hc["events/image"])
        assert is_properly_compressed(hc["logs/test_log"])


def test_copy_logs_variable_length_string():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # add variable-length log to source file
    with h5py.File(path, "a") as h5:
        data = ["hans", "peter", "und", "zapadust"]
        h5.require_group("logs")
        h5["logs"]["var_log"] = data
        assert h5["logs"]["var_log"].dtype.kind == "O"
        assert h5["logs/var_log"].dtype.str == "|O"

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["events/deform"])
        assert is_properly_compressed(hc["events/image"])
        assert is_properly_compressed(hc["logs/var_log"])
        assert hc["logs/var_log"].dtype.kind == "S"
        assert hc["logs/var_log"].dtype.str == "|S100"


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_copy_metadata_config():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    with dclab.new_dataset(path_copy) as ds:
        assert ds.config["experiment"]["sample"] == "background image example"


def test_copy_metadata_of_datasets():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # add log to source file
    with RTDCWriter(path, mode="append") as hw:
        hw.store_log("test_log", ["hans", "peter", "und", "zapadust"])
        hw.h5file["logs/test_log"].attrs["saint"] = "germain"

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["logs/test_log"])
        assert hc["logs/test_log"].attrs["saint"] == "germain"


def test_copy_no_events():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_in = path.with_name("input.rtdc")
    path_copy = path.with_name("test_copy.rtdc")

    with RTDCWriter(path_in, mode="append") as hw, h5py.File(path) as h5:
        hw.h5file.attrs.update(h5.attrs)
        hw.store_log("test_log", ["hans", "peter", "und", "zapadust"])

    # copy
    with h5py.File(path_in) as h5, h5py.File(path_copy, "w") as hc:
        # make sure no events are there
        if "events" in h5:
            assert len(h5["events"]) == 0
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        if "events" in hc:
            assert len(hc["events"]) == 0
        assert "test_log" in hc["logs"]


def test_copy_scalar_features_only():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        # make sure image data is there
        assert "image" in h5["events"]
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc,
                  features="scalar")

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert "image" not in hc["events"]
        assert "deform" in hc["events"]


def test_copy_specified_feature_list():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        # make sure image data is there
        assert "image" in h5["events"]
        assert "area_um" in h5["events"]
        assert "deform" in h5["events"]
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc,
                  features=["image", "deform"])

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert "image" in hc["events"]
        assert "area_um" not in hc["events"]
        assert "deform" in hc["events"]


def test_copy_tables():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # generate a table
    columns = ["bread", "beer", "chocolate"]
    ds_dt = np.dtype({'names': columns,
                      'formats': [np.float64] * len(columns)})
    tab_data = np.zeros((10, len(columns)))
    tab_data[:, 0] = np.arange(10)
    tab_data[:, 1] = 1000
    tab_data[:, 2] = np.linspace(np.pi, 2 * np.pi, 10)
    rec_arr = np.rec.array(tab_data, dtype=ds_dt)
    # sanity check
    assert np.all(rec_arr["bread"][:].flatten() == np.arange(10))
    assert np.all(rec_arr["beer"][:].flatten() == 1000)
    assert np.all(rec_arr["chocolate"][:].flatten() == np.linspace(
        np.pi, 2 * np.pi, 10))

    # add table to source file
    with h5py.File(path, "a") as h5:
        h5tab = h5.require_group("tables")
        h5tab.create_dataset(name="most_important_things",
                             data=rec_arr)
        assert not is_properly_compressed(h5["tables/most_important_things"])

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["tables/most_important_things"])
        tab_data = hc["tables/most_important_things"]
        assert np.all(tab_data["bread"][:].flatten() == np.arange(10))
        assert np.all(tab_data["beer"][:].flatten() == 1000)
        assert np.all(tab_data["chocolate"][:].flatten() == np.linspace(
            np.pi, 2 * np.pi, 10))


def test_copy_tables_array_only():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # generate a table that consists of an array, not a dict-like object
    tab_data = np.random.random((1000, 300))

    # add table to source file
    with h5py.File(path, "a") as h5:
        h5tab = h5.require_group("tables")
        h5tab.create_dataset(name="random_data",
                             data=tab_data)
        assert not is_properly_compressed(h5["tables/random_data"])

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["tables/random_data"])
        assert np.all(tab_data == hc["tables/random_data"])


def test_copy_tables_hdf5_issue_3214():
    """Checks for a bug in HDF5

    https://github.com/HDFGroup/hdf5/issues/3214
    """
    path = retrieve_data("fmt-hdf5_segfault-compound_2023.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # This caused a segmentation fault when using h5py.h5o.copy
    # with "tables/cytoshot_monitor".
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Also make sure metadata are copied
    with h5py.File(path_copy) as hc:
        assert hc["tables"]["cytoshot_monitor"].attrs["COLOR_shift"]\
               == "#0e8f69"


def test_copy_with_compression():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc)

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert is_properly_compressed(hc["events/deform"])
        assert is_properly_compressed(hc["events/image"])


def test_do_not_copy_features():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        # make sure image data is there
        assert "image" in h5["events"]
        assert "deform" in h5["events"]
        rtdc_copy(src_h5file=h5,
                  dst_h5file=hc,
                  features="none")

    # Make sure this worked
    with h5py.File(path_copy) as hc:
        assert "events" not in hc


def test_copier_with_wrong_feature():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")
    path_copy = path.with_name("test_copy.rtdc")

    # copy
    with h5py.File(path) as h5, h5py.File(path_copy, "w") as hc:
        # make sure image data is there
        assert "image" in h5["events"]
        assert "deform" in h5["events"]
        with pytest.raises(ValueError,
                           match="must be either a list of feature names"):
            rtdc_copy(src_h5file=h5,
                      dst_h5file=hc,
                      features="invalid")
        assert "events" not in hc



"""Check RT-DC datasets for completeness"""
import copy
import functools
import warnings

import h5py
import numpy as np

from .copier import is_properly_compressed
from .core import RTDCBase
from .fmt_hdf5 import RTDC_HDF5
from .fmt_hierarchy import RTDC_Hierarchy
from .load import load_file

from .. import definitions as dfn

#: These sections should be fully present, except for the
#: keys in :data:`OPTIONAL_KEYS`.
DESIRABLE_SECTIONS = {
    "experiment",
    "imaging",
    "setup",
}

#: log names that end with these strings are not checked
IGNORED_LOG_NAMES = {
    "_para.ini",
    "_image.ini",
    "FG_Config.mcf",
    "parameters.txt",
    "_SoftwareSettings.ini",
    "dckit-history",
}

#: keys that must be present for every measurement
IMPORTANT_KEYS = {
    "experiment": [
        "date",
        "event count",
        "run index",
        "sample",
        "time"],
    "imaging": [
        "flash device",
        "flash duration",
        "frame rate",
        "pixel size",
        "roi position x",
        "roi position y",
        "roi size x",
        "roi size y"],
    "setup": [
        "channel width",
        "chip region",
        "flow rate",
        "medium"],
}

#: keys that must be present for fluorescence measurements
IMPORTANT_KEYS_FL = {
    "fluorescence": [
        "bit depth",
        "channel count",
        "channels installed",
        "laser count",
        "lasers installed",
        "sample rate",
        "samples per event",
        "signal max",
        "signal min",
        "trace median"],
}

#: maximum line length in log files
LOG_MAX_LINE_LENGTH = 100

#: keys that are optional
OPTIONAL_KEYS = {
    "experiment": [
        "run identifier",
        "timestamp",
    ],
    "fluorescence": [
        "baseline 1 offset",
        "baseline 2 offset",
        "baseline 3 offset",
        # name, lambda, power have their own special tests
        "channel 1 name",
        "channel 2 name",
        "channel 3 name",
        "laser 1 lambda",
        "laser 2 lambda",
        "laser 3 lambda",
        "laser 1 power",
        "laser 2 power",
        "laser 3 power",
    ],
    "setup": [
        "temperature",
        "chip identifier",
    ],
}

#: valid metadata choices
#: .. versionchanged:: 0.29.1
#:    medium not restricted to certain set of choices anymore
VALID_CHOICES = {}


@functools.total_ordering
class ICue(object):
    def __init__(self, msg, level, category, data=None, identifier=None,
                 cfg_section=None, cfg_key=None, cfg_choices=None):
        """Integrity cue"""
        #: human-readable message
        self.msg = msg
        #: severity level ("violation", "alert", or "info")
        self.level = level
        #: machine-readable data associated with the check
        self.data = data
        #: fail category
        self.category = category
        #: identifier e.g. for UI manipulation in DCKit
        self.identifier = identifier
        #: section (only for categories "missing metadata", "bad metadata")
        self.cfg_section = cfg_section
        #: key (only for categories "missing metadata", "bad metadata");
        #: can be omitted e.g. to communicate that the entire section is
        #: missing
        self.cfg_key = cfg_key
        #: allowed choices for the specific [section]: key combination
        #: (only for categories "missing metadata", "bad metadata")
        self.cfg_choices = cfg_choices
        if self.cfg_choices is None:
            if (cfg_section in VALID_CHOICES
                    and cfg_key in VALID_CHOICES[cfg_section]):
                self.cfg_choices = VALID_CHOICES[cfg_section][cfg_key]

    def __eq__(self, other):
        leveld = {"info": 0,
                  "violation": 1,
                  "alert": 2,
                  }
        return ((leveld[self.level], self.cfg_section or "",
                 self.cfg_key or "", self.category, self.msg) ==
                (leveld[other.level], other.cfg_section or "",
                 other.cfg_key or "", other.category, other.msg))

    def __lt__(self, other):
        leveld = {"info": 0,
                  "violation": 1,
                  "alert": 2, }
        return ((leveld[self.level], self.cfg_section or "",
                 self.cfg_key or "", self.category, self.msg) <
                (leveld[other.level], other.cfg_section or "",
                 other.cfg_key or "", other.category, other.msg))

    def __repr__(self):
        return f"<ICue: '{self.msg}' at 0x{hex(id(self))}>"

    @staticmethod
    def get_level_summary(cues):
        """For a list of ICue, return the abundance of all levels"""
        levels = {"info": 0,
                  "alert": 0,
                  "violation": 0}
        for cue in cues:
            levels[cue.level] += 1
        return levels


class IntegrityChecker(object):
    def __init__(self, path_or_ds):
        """Check the integrity of a dataset

        The argument must be either a path to an .rtdc or .tdms file
        or an instance of `RTDCBase`. If a path is given, then all
        warnings (e.g. UnknownConfigurationKeyWarning) are catched
        and added to the cue list.

        Usage:

        .. code:: python

            ic = IntegrityChecker("/path/to/data.rtdc")
            cues = ic.check()

        """
        self.warn_cues = []
        if isinstance(path_or_ds, RTDCBase):
            self.ds = path_or_ds
            self.finally_close = False
        else:
            with warnings.catch_warnings(record=True) as ws:
                warnings.simplefilter("always")
                self.ds = load_file(path_or_ds, enable_basins=False)
                for ww in ws:
                    self.warn_cues.append(ICue(
                        msg=f"{ww.category.__name__}: {ww.message}",
                        level="alert",
                        category="warning"))
            self.finally_close = True

    def __enter__(self):
        return self

    def __exit__(self, *args):
        # close the file
        if self.finally_close and hasattr(self.ds, "__exit__"):
            self.ds.__exit__(*args)

    @property
    def has_fluorescence(self):
        if ("fluorescence" in self.ds
                or "fl1_max" in self.ds
                or "fl2_max" in self.ds
                or "fl3_max" in self.ds):
            fl = True
        else:
            fl = False
        return fl

    def check(self, **kwargs):
        """Run all checks

        This method calls all class methods that start with `check_`.
        `kwargs` are passed to all methods. Possible options are:

        - "expand_section" `bool`: add a cue for every missing
          metadata key if a section is missing
        """
        cues = []
        funcs = IntegrityChecker.__dict__
        if not len(self.ds) == np.sum(self.ds.filter.all):
            raise NotImplementedError(
                "Integrity checks for datasets with active event filters "
                "are not supported!")
        elif self.ds.__class__ == RTDC_Hierarchy:
            raise NotImplementedError(
                "Integrity checks for 'RTDC_Hierarchy' instances are "
                "not supported!")
        for ff in sorted(funcs.keys()):
            if ff.startswith("check_fl_") and not self.has_fluorescence:
                # skip
                continue
            elif ff.startswith("check_"):
                cues += funcs[ff](self, **kwargs)
        return sorted(self.warn_cues + cues)

    def check_basin_features_internal(self, **kwargs):
        """Check whether internal basin features are properly defined"""
        cues = []
        basins = self.ds.basins_get_dicts()
        for bn in basins:
            if bn["type"] == "internal":
                bpaths = bn["paths"]
                if bpaths != ["basin_events"]:
                    cues.append(
                        ICue(msg=f"Uncommon internal basin path: {bpaths}",
                             level="alert",
                             category="basin data",
                             ))
                else:
                    if "basin_events" not in self.ds.h5file:
                        cues.append(
                            ICue(msg="Missing internal basin group "
                                     "'basin_events', although an internal "
                                     "basin is defined",
                                 level="violation",
                                 category="basin data",
                                 ))
                    else:
                        for feat in bn["features"]:
                            if feat not in self.ds.h5file["basin_events"]:
                                cues.append(
                                    ICue(msg=f"Missing internal basin "
                                             f"feature {feat}",
                                         level="violation",
                                         category="basin data",
                                         ))
        return cues

    def check_compression(self, **kwargs):
        cues = []
        if self.ds.format == "tdms":
            compression = "None"
            data = {"compressed": 0,
                    "total": 1,
                    "uncompressed": 0,
                    }
        elif self.ds.format == "hdf5":
            def iter_count_compression(h5):
                comp_i = 0
                noco_i = 0
                for key in h5:
                    obj = h5[key]
                    if isinstance(obj, h5py.Dataset):
                        if is_properly_compressed(obj):
                            # data are compressed with at least level 5
                            comp_i += 1
                        else:
                            # no compression
                            noco_i += 1
                    elif isinstance(obj, h5py.Group):
                        coi, noi = iter_count_compression(obj)
                        comp_i += coi
                        noco_i += noi
                    else:
                        raise ValueError(f"Unknown object: {obj}")
                return comp_i, noco_i

            comp, noco = iter_count_compression(self.ds.h5file)
            if noco == 0:
                compression = "All"
            elif comp == 0:
                compression = "None"
            else:
                compression = f"Partial ({comp} of {noco+comp})"
            data = {"compressed": comp,
                    "total": noco + comp,
                    "uncompressed": noco,
                    }
        else:
            compression = "Unknown"
            data = None
        cues.append(ICue(
            msg=f"Compression: {compression}",
            level="info",
            category="general",
            data=data))
        return cues

    def check_empty(self, **kwargs):
        """The dataset should contain events"""
        cues = []
        lends = len(self.ds)
        if lends == 0:
            cues.append(ICue(
                msg="The dataset does not contain any events",
                level="alert",
                category="feature data"))
        return cues

    def check_external_links(self, **kwargs):
        """An HDF5 dataset should not contain external links"""
        cues = []
        if isinstance(self.ds, RTDC_HDF5):
            has_external, h5object = hdf5_has_external(self.ds.h5file)
            if has_external:
                cues.append(ICue(
                    msg=f"The HDF5 file contains at least one external "
                        f"link: '{h5object}'",
                    level="violation",
                    category="format HDF5"))
        return cues

    def check_feat_index(self, **kwargs):
        """The index of the dataset should be monotonous"""
        cues = []
        lends = len(self.ds)
        if "index" in self.ds:
            if not np.all(self.ds["index"] == np.arange(1, lends + 1)):
                cues.append(ICue(
                    msg="The index feature is not enumerated correctly",
                    level="violation",
                    category="feature data"))
        return cues

    def check_feature_size(self, **kwargs):
        cues = []
        lends = len(self.ds)
        for feat in self.ds.features_innate:
            if feat == "trace":
                for tr in list(self.ds["trace"].keys()):
                    if len(self.ds["trace"][tr]) != lends:
                        cues.append(ICue(
                            msg=f"Features: wrong event count: 'trace/{tr}' "
                                + f"({len(self.ds['trace'][tr])} of {lends})",
                            level="violation",
                            category="feature size"))
            else:
                if len(self.ds[feat]) != lends:
                    cues.append(ICue(
                        msg=f"Features: wrong event count: '{feat}' "
                            + f"({len(self.ds[feat])} of {lends})",
                        level="violation",
                        category="feature size"))
        return cues

    def check_features_unknown_hdf5(self, **kwargs):
        """Check for features that are not defined in dclab

        The idea here is to make sure that third-party software is made
        aware of the fact that it is storing unknown features in an HDF5
        file. If this is the case, then this will e.g. raise an error in
        DCOR-Aid, such that people don't lose new features upon upload.
        """
        # Ignored unknown features are features that are actually known
        # but omitting them is not a big deal.
        ignore_unknown_features = [
            "def",  # An old Shape-In version stored "def" instead of "deform"
            ]
        cues = []
        if self.ds.format == "hdf5":
            for feat in self.ds.h5file["events"]:
                if not dfn.feature_exists(feat):
                    if feat in ignore_unknown_features:
                        continue
                    cues.append(ICue(
                        msg=f"Features: Unknown key '{feat}'",
                        level="violation",
                        category="feature unknown"))
        return cues

    def check_fl_metadata_channel_names(self, **kwargs):
        cues = []
        for ii in range(1, 4):
            chn = f"channel {ii} name"
            fli = f"fl{ii}_max"
            if (fli in self.ds
                    and chn not in self.ds.config["fluorescence"]):
                # Channel names must be defined when there is
                # a corresponding fluorescence signal.
                cues.append(ICue(
                    msg=f"Metadata: Missing key [fluorescence] '{chn}'",
                    level="alert",
                    category="metadata missing",
                    cfg_section="fluorescence",
                    cfg_key=chn))
            elif (fli not in self.ds
                  and chn in self.ds.config["fluorescence"]):
                # Channel names must not be defined when there is
                # no corresponding fluorescence signal.
                cues.append(ICue(
                    msg=f"Metadata: Unused key defined [fluorescence] '{chn}'",
                    level="alert",
                    category="metadata invalid",
                    cfg_section="fluorescence",
                    cfg_key=chn))
        return cues

    def check_fl_num_channels(self, **kwargs):
        cues = []
        # check for number of channels
        if "channel count" in self.ds.config["fluorescence"]:
            chc1 = self.ds.config["fluorescence"]["channel count"]
            chc2 = 0
            for ii in range(1, 4):
                chn = f"channel {ii} name"
                ecn = f"fl{ii}_max"
                if (chn in self.ds.config["fluorescence"] and
                        ecn in self.ds._events):
                    chc2 += 1
            if chc1 != chc2:
                cues.append(ICue(
                    msg="Metadata: fluorescence channel count inconsistent",
                    level="violation",
                    category="metadata wrong",
                    cfg_section="fluorescence",
                    cfg_key="channel count"))
        return cues

    def check_fl_num_lasers(self, **kwargs):
        cues = []
        # check for number of lasers
        if "laser count" in self.ds.config["fluorescence"]:
            lsc1 = self.ds.config["fluorescence"]["laser count"]
            lsc2 = 0
            for ii in range(1, 4):
                kl = f"laser {ii} lambda"
                kp = f"laser {ii} power"
                if (kl in self.ds.config["fluorescence"] and
                        kp in self.ds.config["fluorescence"] and
                        self.ds.config["fluorescence"][kp] != 0):
                    lsc2 += 1
            if lsc1 != lsc2:
                cues.append(ICue(
                    msg="Metadata: fluorescence laser count inconsistent",
                    level="violation",
                    category="metadata wrong",
                    cfg_section="fluorescence",
                    cfg_key="laser count"))
        return cues

    def check_fl_samples_per_event(self, **kwargs):
        cues = []
        # check for samples per event
        if "samples per event" in self.ds.config["fluorescence"]:
            spe = self.ds.config["fluorescence"]["samples per event"]
            if "trace" in self.ds:
                for key in self.ds["trace"].keys():
                    spek = self.ds["trace"][key][0].size
                    if spek != spe:
                        cues.append(ICue(
                            msg="Metadata: wrong number of samples per "
                                + f"event: {key} (expected {spe}, got {spek})",
                            level="violation",
                            category="metadata wrong",
                            cfg_section="fluorescence",
                            cfg_key="samples per event"))
        return cues

    def check_fl_max_positive(self, **kwargs):
        """Check if all fl?_max values are >0.1"""
        cues = []
        neg_feats = []
        for fl in ['fl1_max', 'fl2_max', 'fl3_max']:
            if fl in self.ds:
                if min(self.ds[fl]) <= 0.1:
                    neg_feats.append(fl)
        if neg_feats:
            cues.append(ICue(
                msg=f"Negative value for feature(s): {', '.join(neg_feats)}",
                level="alert",
                category="feature data"))
        return cues

    def check_fl_max_ctc_positive(self, **kwargs):
        """Check if all fl?_max_ctc values are > 0.1"""
        cues = []
        neg_feats = []
        for fl in ['fl1_max_ctc', 'fl2_max_ctc', 'fl3_max_ctc']:
            if fl in self.ds:
                if min(self.ds[fl]) <= 0.1:
                    neg_feats.append(fl)
        if neg_feats:
            cues.append(ICue(
                msg=f"Negative value for feature(s): {', '.join(neg_feats)}",
                level="alert",
                category="feature data"))
        return cues

    def check_flow_rate(self, **kwargs):
        """Make sure sheath and sample flow rates add up"""
        cues = []
        if ("setup" in self.ds.config
                and "flow rate" in self.ds.config["setup"]
                and "flow rate sample" in self.ds.config["setup"]
                and "flow rate sheath" in self.ds.config["setup"]):
            frsum = self.ds.config["setup"]["flow rate"]
            frsam = self.ds.config["setup"]["flow rate sample"]
            frshe = self.ds.config["setup"]["flow rate sheath"]
            if not np.allclose(frsum, frsam + frshe):
                for k in ["flow rate", "flow rate sheath", "flow rate sample"]:
                    cues.append(ICue(
                        msg="Metadata: Flow rates don't add up (sheath "
                            + f"{frshe:g} + sample {frsam:g} "
                            + f"!= channel {frsum:g})",
                        level="alert",
                        category="metadata wrong",
                        cfg_section="setup",
                        cfg_key=k))
        return cues

    def check_fmt_hdf5(self, **kwargs):
        cues = []
        # hdf5-based checks
        if self.ds.format == "hdf5":
            # check meta data of images
            for feat in ["image", "image_bg", "mask"]:
                if feat in self.ds._events:
                    imdat = self.ds[feat]
                    for key, val in [['CLASS', b'IMAGE'],
                                     ['IMAGE_VERSION', b'1.2'],
                                     ['IMAGE_SUBCLASS', b'IMAGE_GRAYSCALE']]:
                        if key not in imdat.attrs:
                            cues.append(ICue(
                                msg=f"HDF5: '/{feat}': missing "
                                    + f"attribute '{key}'",
                                level="alert",
                                category="format HDF5"))
                        elif imdat.attrs.get_id(key).dtype.char != "S":
                            cues.append(ICue(
                                msg=f"HDF5: '/{feat}': attribute '{key}' "
                                    + "should be fixed-length ASCII string",
                                level="alert",
                                category="format HDF5"))
                        elif imdat.attrs[key] != val:
                            cues.append(ICue(
                                msg=f"HDF5: '/{feat}': attribute '{key}' "
                                    + f"should have value '{val}'",
                                level="alert",
                                category="format HDF5"))
            # check length of logs
            with h5py.File(self.ds.path, mode="r") as h5:
                if "logs" in h5:
                    logs = h5["logs"]
                    for logname in logs.keys():
                        # ignore tmds meta data log files
                        lign = [logname.endswith(n) for n in IGNORED_LOG_NAMES]
                        if sum(lign):
                            continue
                        log = logs[logname]
                        for ii in range(len(log)):
                            if len(log[ii]) > LOG_MAX_LINE_LENGTH:
                                cues.append(ICue(
                                    msg=f"Logs: {logname} line {ii} "
                                        + "exceeds maximum line length "
                                        + f"{LOG_MAX_LINE_LENGTH}",
                                    level="alert",
                                    category="format HDF5"))
        return cues

    def check_info(self, **kwargs):
        cues = [
            ICue(
                msg=f"Fluorescence: {self.has_fluorescence}",
                level="info",
                category="general"),
            ICue(
                msg=f"Data file format: {self.ds.format}",
                level="info",
                category="general")]
        return cues

    def check_metadata_bad(self, **kwargs):
        cues = []
        # check for ROI size
        if ("imaging" in self.ds.config
                and "roi size x" in self.ds.config["imaging"]
                and "roi size y" in self.ds.config["imaging"]):
            for ii, roi in enumerate(["roi size y", "roi size x"]):
                for feat in ["image", "image_bg", "mask"]:
                    if feat in self.ds:
                        soll = self.ds[feat].shape[ii+1]
                        ist = self.ds.config["imaging"][roi]
                        if soll != ist:
                            cues.append(ICue(
                                msg="Metadata: Mismatch [imaging] "
                                    + f"'{roi}' and feature {feat} "
                                    + f"({ist} vs {soll})",
                                level="violation",
                                category="metadata wrong",
                                cfg_section="imaging",
                                cfg_key=roi))
        return cues

    def check_metadata_bad_greater_zero(self, **kwargs):
        cues = []
        # check for ROI size
        for sec, key in [
            ["imaging", "frame rate"],
            ["imaging", "pixel size"],
            ["setup", "channel width"],
            ["setup", "flow rate"],
        ]:
            value = self.ds.config.get(sec, {}).get(key)
            if value is not None and value <= 0:
                cues.append(ICue(
                    msg=f"Metadata: Invalid value for [{sec}] '{key}': "
                        + f"'{value}'!",
                    level="violation",
                    category="metadata wrong",
                    cfg_section=sec,
                    cfg_key=key))
        return cues

    def check_metadata_choices(self, **kwargs):
        cues = []
        for sec in VALID_CHOICES:
            for key in VALID_CHOICES[sec]:
                if sec in self.ds.config and key in self.ds.config[sec]:
                    val = self.ds.config[sec][key]
                    if val not in VALID_CHOICES[sec][key]:
                        cues.append(ICue(
                            msg=f"Metadata: Invalid value [{sec}] {key}: "
                                + f"'{val}'",
                            level="violation",
                            category="metadata wrong",
                            cfg_section=sec,
                            cfg_key=key,
                            cfg_choices=VALID_CHOICES[sec][key]))
        return cues

    def check_metadata_hdf5_type(self, **kwargs):
        """This is a low-level HDF5 check"""
        cues = []
        if self.ds.format == "hdf5":
            for entry in self.ds.h5file.attrs:
                if entry.count(":"):
                    sec, key = entry.split(":")
                    val_act = self.ds.h5file.attrs[entry]  # actual value
                    if isinstance(val_act, bytes):
                        val_act = val_act.decode("utf-8")
                    # Check whether the config key exists
                    if dfn.config_key_exists(sec, key):
                        func = dfn.get_config_value_func(sec, key)
                        val_exp = func(val_act)  # expected value
                        if (isinstance(val_exp, (list, tuple, np.ndarray))
                                and np.allclose(val_exp, val_act)):
                            continue
                        elif val_exp == val_act:
                            continue
                        else:
                            cues.append(ICue(
                                msg=f"Metadata: [{sec}]: '{key}' should be "
                                    + f"'{val_exp}', but is '{val_act}'!",
                                level="alert",
                                category="metadata wrong",
                                cfg_section=sec,
                                cfg_key=key))
        return cues

    def check_metadata_online_filter_polygon_points_shape(self, **kwargs):
        cues = []
        if "online_filter" in self.ds.config:
            for key in self.ds.config["online_filter"].keys():
                if key.endswith("polygon points"):
                    points = self.ds.config["online_filter"][key]
                    if points.shape[1] != 2 or points.shape[0] < 3:
                        cues.append(ICue(
                            msg="Metadata: Wrong shape [online_filter] "
                                + f"{key}: '{points.shape}'",
                            level="violation",
                            category="metadata wrong",
                            cfg_section="online_filter",
                            cfg_key=key))
        return cues

    def check_metadata_missing(self, expand_section=True, **kwargs):
        cues = []
        # These "must" be present:
        important = copy.deepcopy(IMPORTANT_KEYS)
        if self.has_fluorescence:
            important.update(IMPORTANT_KEYS_FL)
        # A list of sections we would like to investigate
        secs_investiage = list(set(important.keys()) | set(DESIRABLE_SECTIONS))

        for sec in secs_investiage:
            if sec not in self.ds.config and not expand_section:
                cues.append(ICue(
                    msg=f"Metadata: Missing section '{sec}'",
                    level="violation" if sec in important else "alert",
                    category="metadata missing",
                    cfg_section=sec))
            else:
                for key in dfn.config_keys[sec]:
                    if key not in self.ds.config[sec]:
                        if sec in OPTIONAL_KEYS and key in OPTIONAL_KEYS[sec]:
                            # ignore this key
                            continue
                        elif sec in important and key in important[sec]:
                            level = "violation"
                        else:
                            level = "alert"
                        cues.append(ICue(
                            msg=f"Metadata: Missing key [{sec}] '{key}'",
                            level=level,
                            category="metadata missing",
                            cfg_section=sec,
                            cfg_key=key))
        # check for temperature
        if "temp" in self.ds:
            if "temperature" not in self.ds.config["setup"]:
                cues.append(ICue(
                    msg="Metadata: Missing key [setup] 'temperature', "
                        + "because the 'temp' feature is given",
                    level="alert",
                    category="metadata missing",
                    cfg_section="setup",
                    cfg_key="temperature"))
        return cues

    def check_ml_class(self, **kwargs):
        """Try to comput ml_class feature and display error message"""
        cues = []
        if "ml_class" in self.ds:
            try:
                self.ds["ml_class"]
            except ValueError as e:
                cues.append(ICue(
                    msg=e.args[0],
                    level="violation",
                    category="feature data"))
        return cues

    def check_shapein_issue3_bad_medium(self, **kwargs):
        """Some versions of Shape-In stored wrong [setup]: medium

        The problem only affects selection of "CellCarrier" which had
        index 0 and as a result "CellCarrierB" was written to the file.
        This means we only have to check for the Shape-In version and
        whether the medium is 'CellCarrierB'. In DCKit, the user can
        then manually edit the medium.

        Affected Shape-In versions: >=2.2.1.0,<2.2.2.3
        Also affected if bad config file was used: 2.2.2.4, 2.3.0.0

        https://github.com/ZELLMECHANIK-DRESDEN/ShapeIn_Issues/issues/3
        """
        cues = []
        medium = self.ds.config["setup"].get("medium", "")
        si_ver = self.ds.config["setup"].get("software version", "")
        si_ver = si_ver.strip("dev")  # for e.g. "2.2.1.0dev"
        if (medium == "CellCarrierB"
                and si_ver in ["2.2.1.0", "2.2.2.0", "2.2.2.1", "2.2.2.2",
                               "2.2.2.3",
                               # The issue can still occur for these versions
                               # of Shape-In if a bad configuration file
                               # is used.
                               "2.2.2.4", "2.3.0.0"]):
            cues.append(ICue(
                msg="Metadata: Please verify that 'medium' is really "
                    + "'CellCarrierB' (Shape-In issue #3)",
                level="alert",
                category="metadata wrong",
                identifier="Shape-In issue #3",
                cfg_section="setup",
                cfg_key="medium"))
        return cues

    def check_temperature_zero_zmd(self, **kwargs):
        """If there is a loose cable, then temperature might be all-zero

        https://github.com/DC-analysis/dclab/issues/183
        """
        cues = []
        from_zmd = "ZMD" in self.ds.config["setup"].get("identifier", "")
        if from_zmd and "temp" in self.ds:
            temp = self.ds["temp"]
            if np.allclose(temp[:10], 0) and np.allclose(temp, 0):
                cues.append(ICue(
                    msg="Feature: The 'temp' feature is all-zero, check "
                        "the cables of the temperature sensor!",
                    level="violation",
                    category="feature data"))
        return cues

    def sanity_check(self):
        """Sanity check that tests whether the data can be accessed"""
        cues = []
        cues += self.check_feature_size()
        cues += self.check_metadata_bad()
        for cue in cues:
            cue.msg = "Sanity check failed: " + cue.msg
        return [ci for ci in cues if ci.level == "violation"]


def check_dataset(path_or_ds):
    """Check whether a dataset is complete

    Parameters
    ----------
    path_or_ds: str or pathlib.Path or RTDCBase
        Full path to a dataset on disk or an instance of RTDCBase

    Returns
    -------
    violations: list of str
        Dataset format violations (hard)
    alerts: list of str
        Dataset format alerts (soft)
    info: list of str
        Dataset information
    """
    aler = []
    info = []
    viol = []
    with IntegrityChecker(path_or_ds) as ic:
        # perform all checks
        icues = ic.check(expand_section=False)
        for cue in icues:
            if cue.level == "info":
                info.append(cue.msg)
            elif cue.level == "alert":
                aler.append(cue.msg)
            elif cue.level == "violation":
                viol.append(cue.msg)
    return sorted(viol), sorted(aler), sorted(info)


def hdf5_has_external(h5):
    """Check recursively, whether an h5py object contains external data

    External data includes binary data in external files, virtual
    datasets, and external links.

    Returns a tuple of either

    - `(True, path_ext)` if the object contains external data
    - `(False, None)` if this is not the case

    where `path_ext` is the path to the group or dataset in `h5`.

    .. versionadded:: 0.62.0

    """
    for key in h5:
        obj = h5[key]
        if (obj.file != h5.file  # not in same file
                or (isinstance(obj, h5py.Dataset)
                    and (obj.is_virtual  # virtual dataset
                         or obj.external))):  # external dataset
            # These are external data
            return True, f"{h5.name}/{key}".replace("//", "/")
        elif isinstance(obj, h5py.Group):
            # Perform recursive check for external data
            has_ext, path_ext = hdf5_has_external(obj)
            if has_ext:
                return True, path_ext
    else:
        return False, None



from ... import features
from .ancillary_feature import AncillaryFeature


def compute_contour(mm):
    cont = features.contour.get_contour_lazily(mask=mm["mask"])
    return cont


def compute_bright(mm):
    bavg, bsd = features.bright.get_bright(
        mask=mm["mask"],
        image=mm["image"],
        ret_data="avg,sd",
        )
    return {"bright_avg": bavg, "bright_sd": bsd}


def compute_bright_bc(mm):
    bavg, bsd = features.bright_bc.get_bright_bc(
        mask=mm["mask"],
        image=mm["image"],
        image_bg=mm["image_bg"],
        bg_off=mm["bg_off"] if "bg_off" in mm else None,
        ret_data="avg,sd",
        )
    return {"bright_bc_avg": bavg, "bright_bc_sd": bsd}


def compute_bright_perc(mm):
    p10, p90 = features.bright_perc.get_bright_perc(
        mask=mm["mask"],
        image=mm["image"],
        image_bg=mm["image_bg"],
        bg_off=mm["bg_off"] if "bg_off" in mm else None,
        )
    return {"bright_perc_10": p10, "bright_perc_90": p90}


def compute_inert_ratio_cvx(mm):
    return features.inert_ratio.get_inert_ratio_cvx(cont=mm["contour"])


def compute_inert_ratio_prnc(mm):
    return features.inert_ratio.get_inert_ratio_prnc(cont=mm["contour"])


def compute_inert_ratio_raw(mm):
    return features.inert_ratio.get_inert_ratio_raw(cont=mm["contour"])


def compute_tilt(mm):
    return features.inert_ratio.get_tilt(cont=mm["contour"])


def compute_volume(mm):
    vol = features.volume.get_volume(
        cont=mm["contour"],
        pos_x=mm["pos_x"],
        pos_y=mm["pos_y"],
        pix=mm.config["imaging"]["pixel size"])
    return vol


def register():
    AncillaryFeature(feature_name="contour",
                     method=compute_contour,
                     req_features=["mask"])

    AncillaryFeature(feature_name="bright_avg",
                     method=compute_bright,
                     req_features=["image", "mask"])

    AncillaryFeature(feature_name="bright_sd",
                     method=compute_bright,
                     req_features=["image", "mask"])

    AncillaryFeature(feature_name="bright_bc_avg",
                     method=compute_bright_bc,
                     req_features=["image", "image_bg", "mask"])

    AncillaryFeature(feature_name="bright_bc_sd",
                     method=compute_bright_bc,
                     req_features=["image", "image_bg", "mask"])

    AncillaryFeature(feature_name="bright_perc_10",
                     method=compute_bright_perc,
                     req_features=["image", "image_bg", "mask"])

    AncillaryFeature(feature_name="bright_perc_90",
                     method=compute_bright_perc,
                     req_features=["image", "image_bg", "mask"])

    AncillaryFeature(feature_name="inert_ratio_cvx",
                     method=compute_inert_ratio_cvx,
                     req_features=["contour"])

    AncillaryFeature(feature_name="inert_ratio_prnc",
                     method=compute_inert_ratio_prnc,
                     req_features=["contour"])

    AncillaryFeature(feature_name="inert_ratio_raw",
                     method=compute_inert_ratio_raw,
                     req_features=["contour"])

    AncillaryFeature(feature_name="tilt",
                     method=compute_tilt,
                     req_features=["contour"])

    AncillaryFeature(feature_name="volume",
                     method=compute_volume,
                     req_features=["contour", "pos_x", "pos_y"],
                     req_config=[["imaging", ["pixel size"]]])



import pathlib
import tempfile

from dclab import new_dataset, util
import h5py
import pytest

from helper_methods import retrieve_data


@pytest.fixture(autouse=True)
def clear_cache():
    util.hashfile.cache_clear()


def test_hashfile_basic():
    td = pathlib.Path(tempfile.mkdtemp("test_hashfile"))
    p1 = td / "test_1.txt"
    p1.write_text("Lorem ipsum")
    util.hashfile(p1)
    util.hashfile(p1)
    assert util.hashfile.cache_info().misses == 1
    assert util.hashfile.cache_info().hits == 1
    assert util.hashfile.cache_info().currsize == 1

    p2 = td / "test_2.txt"
    p2.write_text("dolor sit amet.")
    util.hashfile(p2)
    assert util.hashfile.cache_info().misses == 2
    assert util.hashfile.cache_info().hits == 1
    assert util.hashfile.cache_info().currsize == 2


def test_hashfile_modified():
    td = pathlib.Path(tempfile.mkdtemp("test_hashfile"))
    p1 = td / "test_1.txt"
    p1.write_text("Lorem ipsum")
    util.hashfile(p1)
    util.hashfile(p1)
    assert util.hashfile.cache_info().misses == 1
    assert util.hashfile.cache_info().hits == 1
    assert util.hashfile.cache_info().currsize == 1

    p1.write_text("dolor sit amet.")
    util.hashfile(p1)
    assert util.hashfile.cache_info().misses == 2
    assert util.hashfile.cache_info().hits == 1
    assert util.hashfile.cache_info().currsize == 2


def test_hashfile_modified_quickly():
    td = pathlib.Path(tempfile.mkdtemp("test_hashfile"))
    p1 = td / "test_1.txt"
    p1.write_text("Lorem ipsum")
    util.hashfile(p1)
    p1.write_text("dolor sit amet.")
    util.hashfile(p1)
    assert util.hashfile.cache_info().misses == 2
    assert util.hashfile.cache_info().hits == 0
    assert util.hashfile.cache_info().currsize == 2


def test_hash_hdf5_dataset():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")

    with h5py.File(path, "r") as h5:
        hash1 = util.hashobj(h5["/events/image"])

    # sanity check
    with h5py.File(path, "r") as h5:
        hash2 = util.hashobj(h5["/events/image"])
    assert hash1 == hash2

    # now change something
    with h5py.File(path, "a") as h5:
        h5.attrs["setup:medium"] = "water"

    # actual test
    with h5py.File(path, "r") as h5:
        hash3 = util.hashobj(h5["/events/image"])
    assert hash1 != hash3


@pytest.mark.filterwarnings(
    "ignore::dclab.rtdc_dataset.config.WrongConfigurationTypeWarning")
def test_hash_configuration():
    path = retrieve_data("fmt-hdf5_image-bg_2020.zip")

    with new_dataset(path) as ds:
        hash1 = util.hashobj(ds.config)

    # sanity check (change something unrelated)
    with h5py.File(path, "a") as h5:
        h5["events"].attrs["peter pan"] = "hook"

    with new_dataset(path) as ds:
        hash2 = util.hashobj(ds.config)
    assert hash1 == hash2

    # now the actual test
    with h5py.File(path, "a") as h5:
        h5.attrs["experiment:sample"] = "Iratrax"

    with new_dataset(path) as ds:
        hash3 = util.hashobj(ds.config)

    assert hash1 != hash3
